import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m4_primitives import *

from base_aux.aux_eq.m3_eq_valid3_derivatives import *
from base_aux.base_nest_dunders.m1_init2_annots1_attrs_by_args_kwargs import *


# =====================================================================================================================
def _ExpectAux__eq_in__all_operators(eq_cls: type[Base_EqValid], other, _EXPECTED, *args, **kwargs: dict[str, Any]) -> None | NoReturn:
    """
    GOAL
    ----
    check one EqClass by 4 variant lines
    """
    Lambda(eq_cls(*args, **kwargs) == other).check_expected__assert(_EXPECTED)
    # Lambda(eq_cls(*args, **kwargs, _iresult_reverse=True) == other).check_assert(not _EXPECTED)

    Lambda(other in eq_cls(*args, **kwargs)).check_expected__assert(_EXPECTED)
    # Lambda(other in eq_cls(*args, _iresult_reverse=True)).check_assert(not _EXPECTED)


# ---------------------------------------------------------------------------------------------------------------------
def test__0_EqValidObj__usage_operators():
    assert 1 == EqValid_BoolTrue()
    assert 1 in EqValid_BoolTrue()
    assert 0 not in EqValid_BoolTrue()

    assert 1 in [0, 1]
    assert 1 in [0, EqValid_BoolTrue()]

    assert 0 in [0, EqValid_BoolTrue()]
    assert 0 not in [1, EqValid_BoolTrue()]


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="_validator, other, _EXPECTED",
    argvalues=[
        (bool, None, False),
        (bool, 0, False),
        (bool, 1, True),
        (bool, 2, True),
        (bool, LAMBDA_TRUE, True),
        (bool, LAMBDA_FALSE, False),
    ]
)
def test__0_Base_EqValid__single(_validator, other, _EXPECTED):
    Lambda(Base_EqValid(_validator=_validator) == other).check_expected__assert(_EXPECTED)
    Lambda(Base_EqValid(_validator=_validator, _iresult_reverse=True) == other).check_expected__assert(not _EXPECTED)

    Lambda(other in Base_EqValid(_validator=_validator)).check_expected__assert(_EXPECTED)
    Lambda(other in Base_EqValid(_validator=_validator, _iresult_reverse=True)).check_expected__assert(not _EXPECTED)


# ---------------------------------------------------------------------------------------------------------------------
def _victim_validator(source, arg) -> bool:
    return source == arg


@pytest.mark.parametrize(
    argnames="_validator, other, args, _EXPECTED",
    argvalues=[
        # ALL True ----
        (_victim_validator, 1, [1, 1], [True, True, False, False]),
        (_victim_validator, LAMBDA_1, [1, 1], [True, True, False, False]),

        # # ALL False ----
        (_victim_validator, 0, [1, 1], [False, False, True, True]),
        (_victim_validator, LAMBDA_0, [1, 1], [False, False, True, True]),
    ]
)
def test__0_Base_EqValid__args_cumulate_1__ALL(_validator, other, args, _EXPECTED):
    for index, _iresult_cumulate in enumerate(EnumAdj_BoolCumulate):
        print()
        print(index, _iresult_cumulate)
        Lambda(Base_EqValid(*args, _validator=_validator, _iresult_cumulate=_iresult_cumulate) == other).check_expected__assert(_EXPECTED[index])
        Lambda(Base_EqValid(*args, _validator=_validator, _iresult_cumulate=_iresult_cumulate, _iresult_reverse=True) == other).check_expected__assert(not _EXPECTED[index])

        Lambda(other in Base_EqValid(*args, _validator=_validator, _iresult_cumulate=_iresult_cumulate)).check_expected__assert(_EXPECTED[index])
        Lambda(other in Base_EqValid(*args, _validator=_validator, _iresult_cumulate=_iresult_cumulate, _iresult_reverse=True)).check_expected__assert(not _EXPECTED[index])


@pytest.mark.parametrize(
    argnames="_validator, other, args, _iresult_cumulate, _EXPECTED",
    argvalues=[
        # [direct, reverse]
        # ALL True ----
        (_victim_validator, 1, [1, 1], EnumAdj_BoolCumulate.ANY_TRUE, [True, False]),
        (_victim_validator, 1, [1, 1], EnumAdj_BoolCumulate.ANY_FALSE, [False, True]),

        # # ALL False ----
        (_victim_validator, 0, [1, 1], EnumAdj_BoolCumulate.ANY_TRUE, [False, True]),
        (_victim_validator, 0, [1, 1], EnumAdj_BoolCumulate.ANY_FALSE, [True, False]),

        # ANY
        (_victim_validator, 1, [0, 1], EnumAdj_BoolCumulate.ANY_TRUE, [True, True]),
        (_victim_validator, 1, [0, 1], EnumAdj_BoolCumulate.ANY_FALSE, [True, True]),
    ]
)
def test__0_Base_EqValid__args_cumulate_2__ANY(_validator, other, args, _iresult_cumulate, _EXPECTED):
    Lambda(Base_EqValid(*args, _validator=_validator, _iresult_cumulate=_iresult_cumulate) == other).check_expected__assert(_EXPECTED[0])
    Lambda(Base_EqValid(*args, _validator=_validator, _iresult_cumulate=_iresult_cumulate, _iresult_reverse=True) == other).check_expected__assert(_EXPECTED[1])

    Lambda(other in Base_EqValid(*args, _validator=_validator, _iresult_cumulate=_iresult_cumulate)).check_expected__assert(_EXPECTED[0])
    Lambda(other in Base_EqValid(*args, _validator=_validator, _iresult_cumulate=_iresult_cumulate, _iresult_reverse=True)).check_expected__assert(_EXPECTED[1])


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, other, _EXPECTED",
    argvalues=[
        ((bool, int), True, True),
        ((bool, int), 1, True),
        ((bool, int), 1.0, False),

        ((bool, int, None), 1, True),
        ((1, 2), 1, True),
    ]
)
def test__isinstance(args, other, _EXPECTED):
    _ExpectAux__eq_in__all_operators(EqValid_IsinstanceSameinstance, other, _EXPECTED, *args)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, other, _EXPECTED",
    argvalues=[
        ([1,2], 1, (False, True)),
        ([1,2], "1", (False, True)),
        ([1,2], 10, (False, True)),
        ([*"12"], "1", (True, True)),
        ([*"12"], "10", (True, True)),
        ([*"12"], "hello", (False, False)),

        ([*"ABC"], "A", (True, True)),
        ([*"ABC"], "a", (False, True)),
        ([*"ABC"], "f", (False, False)),
    ]
)
def test__contains(args, other, _EXPECTED):
    _ExpectAux__eq_in__all_operators(EqValid_Contain, other, _EXPECTED[0],*args)
    _ExpectAux__eq_in__all_operators(EqValid_ContainStrIc, other, _EXPECTED[1], *args)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, other, _EXPECTED",
    argvalues=[
        ([*"ABC"], "A1", (True, True)),
        ([*"ABC"], "a1", (False, True)),
        ([*"ABC"], "f1", (False, False)),
    ]
)
def test__startswith(args, other, _EXPECTED):
    _ExpectAux__eq_in__all_operators(EqValid_Startswith, other, _EXPECTED[0], *args)
    _ExpectAux__eq_in__all_operators(EqValid_StartswithIc, other, _EXPECTED[1], *args)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="args, other, _EXPECTED",
    argvalues=[
        ([*"ABC"], "1A", (True, True)),
        ([*"ABC"], "1a", (False, True)),
        ([*"ABC"], "1f", (False, False)),
    ]
)
def test__endswith(args, other, _EXPECTED):
    _ExpectAux__eq_in__all_operators(EqValid_Endswith, other, _EXPECTED[0], *args)
    _ExpectAux__eq_in__all_operators(EqValid_EndswithIc, other, _EXPECTED[1], *args)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="other, _EXPECTED",
    argvalues=[
        (False, (False, False, False, True, False)),
        (True, (True, False, False, True, False)),
        (1, (True, False, False, True, False)),
        (Exception, (False, True, False, True, True)),
        (Exception(), (False, True, False, True, True)),
        (LAMBDA_EXC, (False, True, False, True, True)),
        (LAMBDA_RAISE, (False, False, True, False, True)),
    ]
)
def test__bool_exc_raise(other, _EXPECTED):
    args = ()
    _ExpectAux__eq_in__all_operators(EqValid_BoolTrue, other, _EXPECTED[0], *args)
    _ExpectAux__eq_in__all_operators(EqValid_Exc, other, _EXPECTED[1], *args)
    _ExpectAux__eq_in__all_operators(EqValid_Raise, other, _EXPECTED[2], *args)
    _ExpectAux__eq_in__all_operators(EqValid_NotRaise, other, _EXPECTED[3], *args)
    _ExpectAux__eq_in__all_operators(EqValid_ExcRaise, other, _EXPECTED[4], *args)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, other, _EXPECTED",
    argvalues=[
        ([1, 2], 1, (True, True, True)),
        ([1, 2], "1", (False, True, True)),
        ([1, 2], 10, (False, False, False)),
        ([*"12"], "1", (True, True, False)),
        ([*"12"], "10", (False, False, False)),
        ([*"12"], "hello", (False, False, False)),

        ([*"ABC"], "A", (True, True, False)),
        ([*"ABC"], "a", (False, True, False)),
        ([*"ABC"], "f", (False, False, False)),
    ]
)
def test__EQ(args, other, _EXPECTED):
    _ExpectAux__eq_in__all_operators(EqValid_EQ, other, _EXPECTED[0], *args)
    _ExpectAux__eq_in__all_operators(EqValid_EQ_StrIc, other, _EXPECTED[1], *args)
    _ExpectAux__eq_in__all_operators(EqValid_EQ_NumParsedSingle, other, _EXPECTED[2], *args)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="args, other, _EXP_obj, _EXP_num",
    argvalues=[
        # FAILS ------
        # (
        #         (1,2),
        #         False,      # be careful! False < 1!!!!
        #         (False, False, False, False),
        #         (False, False, False, False),
        # ),
        (
                (1, 2), Exception,
                (False, False, False, False),
                (False, False, False, False),
        ),
        (
                (Exception, 2), 1,
                (False, False, False, False),
                (False, False, False, False),
        ),
        (
                (1, 2), "hello",
                (False, False, False, False),
                (False, False, False, False),
        ),

        # correct cmp ------
        (
                (1, 2), 0,
                (True, True, False, False),
                (True, True, False, False),
        ),
        # ------------------
        (
                (1, ), 1,
                (False, True, True, False),
                (False, True, True, False),
        ),
        (
                (1, 2), 1,
                (False, True, False, False),
                (False, True, False, False),
        ),

        (
                (1, 2), "a1c",
                (False, False, False, False),
                (False, True, False, False),
        ),

        # ------------------
        (
                (1, ), 2,
                (False, False, True, True),
                (False, False, True, True),
        ),
        (
                (1, 2), 2,
                (False, False, True, False),
                (False, False, True, False),
        ),
        (
                (1, 2), "a2c",
                (False, False, False, False),
                (False, False, True, False),
        ),

        # ------------------
        (
                (1, 2), 3,
                (False, False, True, True),
                (False, False, True, True),
        ),
    ]
)
def test__LGTE_1_variants(args, other, _EXP_obj, _EXP_num):
    # ------
    _ExpectAux__eq_in__all_operators(EqValid_LT, other, _EXP_obj[0], *args)
    _ExpectAux__eq_in__all_operators(EqValid_LE, other, _EXP_obj[1], *args)
    _ExpectAux__eq_in__all_operators(EqValid_GE, other, _EXP_obj[2], *args)
    _ExpectAux__eq_in__all_operators(EqValid_GT, other, _EXP_obj[3], *args)

    # ------
    _ExpectAux__eq_in__all_operators(EqValid_LT_NumParsedSingle, other, _EXP_num[0], *args)
    _ExpectAux__eq_in__all_operators(EqValid_LE_NumParsedSingle, other, _EXP_num[1], *args)
    _ExpectAux__eq_in__all_operators(EqValid_GE_NumParsedSingle, other, _EXP_num[2], *args)
    _ExpectAux__eq_in__all_operators(EqValid_GT_NumParsedSingle, other, _EXP_num[3], *args)

@pytest.mark.parametrize(
    argnames="kwargs, other, _EXPECTED",
    argvalues=[
        # FAILS ------
        # (dict(lt=1, ), Exception, (False, False)),  # be careful! False < 1!!!!
        (dict(lt=1, ), Exception, (False, False)),
        (dict(lt=Exception, ), 1, (False, False)),
        (dict(lt=1, ), "hello", (False, False)),

        # correct cmp ------
        (dict(gt=1, lt=2), 0, (False, False)),
        (dict(gt=1, lt=2), "aaa0bbb", (False, False)),

        # ------------------
        (dict(gt=1, lt=2), 1, (False, False)),
        (dict(ge=1, lt=2), 1, (True, True)),
        (dict(ge=1, lt=2), "aaa1bbb", (False, True)),

        # ------------------
        (dict(gt=1, lt=2), 2, (False, False)),
        (dict(gt=1, le=2), 2, (True, True)),
        (dict(gt=1, le=2), "aaa2bbb", (False, True)),

        # ------------------
        (dict(gt=1, lt=2), 3, (False, False)),
        (dict(gt=1, lt=2), "aaa3bbb", (False, False)),
    ]
)
def test__LGTE_2_dict(kwargs, other, _EXPECTED):
    _ExpectAux__eq_in__all_operators(EqValid_LGTE, other, _EXPECTED[0], **kwargs)
    _ExpectAux__eq_in__all_operators(EqValid_LGTE_NumParsedSingle, other, _EXPECTED[1], **kwargs)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="other, value, _EXPECTED",
    argvalues=[
        # TRASH ----
        (True, True, (False, False, False, False)),
        (True, None, (False, True, False, False)),
        (True, 1, (False, False, False, False)),
        ("", False, (False, False, False, False)),

        # VALUES ----
        ("123", False, (True, False, True, False)),
        ("123", 1, (True, False, True, False)),
        ("123", 123, (True, True, True, False)),
        ("123", "123", (True, False, True, False)),

        ("a123a", "b123b", (True, False, True, False,)),
        ("a123a", 123, (True, True, True, False,)),
        ("a123a", 1, (True, False, True, False,)),

        ("a1.2.3a", "b123b", (False, False, False, False, )),
        ("a1.2.3a", "hello", (False, False, False, False, )),

        ("a1.00a", "b001bb", (True, False, False, True, )),
    ]
)
def test__EqValid_NumParsedSingle(other, value, _EXPECTED):
    _ExpectAux__eq_in__all_operators(EqValid_NumParsedSingle_Success, other, _EXPECTED[0])
    _ExpectAux__eq_in__all_operators(EqValid_NumParsedSingle_EQ, other, _EXPECTED[1], value)
    _ExpectAux__eq_in__all_operators(EqValid_EQ_NumParsedSingle, other, _EXPECTED[1], value)
    _ExpectAux__eq_in__all_operators(EqValid_NumParsedSingle_TypeInt, other, _EXPECTED[2])
    _ExpectAux__eq_in__all_operators(EqValid_NumParsedSingle_TypeFloat, other, _EXPECTED[3])


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="args, other, _EXPECTED",
    argvalues=[
        ((r"\d", ), 1, (True, True, False, False)),
        ((r"\d\d", ), 1, (False, False, True, True)),
        ((r"\d", r"\d\d", ), 1, (False, True, False, True)),

        ((r"\d\d",), LAMBDA_RAISE, (False, False, True, True)),
        ((r"\d\d",), LAMBDA_EXC, (False, False, True, True)),

        ((r"true",), "Tr", (False, False, True, True)),
        ((r"true",), "True", (True, True, False, False)),
        ((r"true",), LAMBDA_TRUE, (True, True, False, False)),
    ]
)
def test__regexp(args, other, _EXPECTED):
    Lambda(EqValid_Regexp(*args, _iresult_cumulate=EnumAdj_BoolCumulate.ALL_TRUE) == other).check_expected__assert(_EXPECTED[0])
    Lambda(EqValid_Regexp(*args, _iresult_cumulate=EnumAdj_BoolCumulate.ANY_TRUE) == other).check_expected__assert(_EXPECTED[1])
    Lambda(EqValid_Regexp(*args, _iresult_cumulate=EnumAdj_BoolCumulate.ALL_FALSE) == other).check_expected__assert(_EXPECTED[2])
    Lambda(EqValid_Regexp(*args, _iresult_cumulate=EnumAdj_BoolCumulate.ANY_FALSE) == other).check_expected__assert(_EXPECTED[3])

    # ---------
    _ExpectAux__eq_in__all_operators(EqValid_RegexpAllTrue, other, _EXPECTED[0], *args)
    _ExpectAux__eq_in__all_operators(EqValid_RegexpAnyTrue, other, _EXPECTED[1], *args)
    _ExpectAux__eq_in__all_operators(EqValid_RegexpAllFalse, other, _EXPECTED[2], *args)
    _ExpectAux__eq_in__all_operators(EqValid_RegexpAnyFalse, other, _EXPECTED[3], *args)


def test__regexp_manual():
    Lambda(EqValid_Regexp(r"\d+[.,]?\d*V") == "11.688889V").check_expected__assert()


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="source, other, _EXPECTED",
    argvalues=[
        # (dict(a1=1), True, False),
        (dict(a1=1), dict(a1=11), False),
        (dict(a1=1), dict(a1=1), True),
        (dict(a1=1), dict(a1=1, b2=2), True),

        (dict(a1=1, a2=1), dict(a1=1), False),
        (dict(a1=1, a2=1), dict(A1=1, A2=1), True),
    ]
)
def test__AttrsByKwargs(source, other, _EXPECTED):
    Lambda(EqValid_AttrsByKwargs(**source) == NestInit_AnnotsAttr_ByArgsKwargs(**other)).check_expected__assert(_EXPECTED)


@pytest.mark.parametrize(
    argnames="source, other, _EXPECTED",
    argvalues=[
        # OBVIOUS ------
        (dict(a1=1), dict(a1=11), (False, False)),
        (dict(a1=1), dict(a1=1), (True, True)),
        (dict(a1=1), dict(a1=1, b2=2), (True, True)),

        (dict(a1=1, a2=1), dict(a1=1), (False, False)),
        (dict(a1=1, a2=1), dict(A1=1, A2=1), (True, True)),

        # MESHED LEVELS ------
        (dict(a1=1, _a2=2), dict(a1=1), (False, True)),
    ]
)
def test__AttrsByObj(source, other, _EXPECTED):
    Lambda(
        EqValid_AttrsByObjNotPrivate(NestInit_AnnotsAttr_ByArgsKwargs(**source)) == NestInit_AnnotsAttr_ByArgsKwargs(**other)
    ).check_expected__assert(_EXPECTED[0])
    Lambda(
        EqValid_AttrsByObjNotHidden(NestInit_AnnotsAttr_ByArgsKwargs(**source)) == NestInit_AnnotsAttr_ByArgsKwargs(**other)
    ).check_expected__assert(_EXPECTED[1])


# =====================================================================================================================



from typing import *
import copy
import re


# =====================================================================================================================
class DictDotAttrAccess(dict):  # STARICHENKO
    """dot.notation access to dictionary attributes.
    first variant!
    CANT CREATE SINGLETON!

    cant create singleton!!! i was trying!!! even by typical def __***

    typical usage:
        my_dict = {'param1': 'GOOD'}
        nested_dict = {'param1': 'nested GOOD too'}
        print(my_dict)  # {'param1': 'GOOD'}

        # EXECUTE
        my_dict = DictDotAttrAccess(my_dict)

        # READING
        print(my_dict)  # {'param1': 'GOOD'}
        print(my_dict.param1)  # 'GOOD'
        print(my_dict["param1"])  # 'GOOD'

        # CHANGE
        my_dict.param1 = 123
        print(my_dict)  # {'param1': 123}
        print(my_dict.param1)  # 123
        my_dict["param1"] = 1000
        print(my_dict)  # {'param1': 1000}
        print(my_dict.param1)  # 1000

        # NESTED
        my_dict.nested = nested_dict
        print(my_dict)  # {'param1': 1000, 'nested': {'param1': 'nested GOOD too'}}
        # print(my_dict.nested.param1)  # AttributeError: 'dict' object has no attribute 'param1'

        my_dict.nested = DictDotAttrAccess(nested_dict)
        print(my_dict)  # {'param1': 1000, 'nested': {'param1': 'nested GOOD too'}}
        print(my_dict.nested.param1)  # 'nested GOOD too'

    ОСОБЕННОСТИ:
        +++И ТАК РАБОТАЛО
            1. +BLANK
                my_dict = DictDotAttrAccess()
                print(my_dict)  # {}

            2. +BOOL
                my_dict = DictDotAttrAccess()
                print(bool(my_dict))  # False

                my_dict = DictDotAttrAccess({'p1': '111'})
                print(my_dict)  # {'p1': '111'}
                print(bool(my_dict))  # True

            3. +IN LIST
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                print("p1" in my_dict)  # True

            5. +при обращении к НЕCУЩЕСТВУЮЩЕМУ - выдает None! без добавления параметра, а после присвоения - параметр создается
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                print(my_dict.p2)  #None
                print(my_dict)  #{'p1': '111'}
                my_dict.p2 = 222
                print(my_dict)  #{'p1': '111', 'p2': 222}

            6. +UPDATE
                my_dict= DictDotAttrAccess()
                print(my_dict)  # {}

                my_dict.update({1:1})
                print(my_dict)  # {1: 1}

            6. +POP
                my_dict = DictDotAttrAccess({1:1})

                my_dict.pop(1)
                print(my_dict)  # {}

            8. +ITEMS()
                my_dict = DictDotAttrAccess({'p1': '111'})
                print(my_dict.items())  #dict_items([('p1', '111')])

        +++ПРИШЛОСЬ ДОБАВИТЬ
            10. +ИТЕРИРОВАТЬ
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                for i in my_dict:
                    print(i, my_dict.get(i))    # p1 111

            11. +КОПИЮ
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                my_dict2 = copy.copy(my_dict)
                print(my_dict2)  # {'p1': '111'}


        ---НЕ ПОЛУЧИЛОСЬ РЕАЛИЗОВАТЬ
            20. замену для __MISSING__
                def __missing__(self, key):
                    return "HELLO"

                __missing__ = lambda: "HELLO"
                __missing__ = FUNC_LINK_LAMBDA_REPEATER
    """
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
    __iter__ = dict.__iter__
    __copy__ = dict.copy

    # __repr__ = dict.__repr__  # и так работает!
    # __str__ = dict.__str__    # и так работает!
    # __len__ = dict.__len__    # и так работает!


# dict-FIND
def dict_find_pattern(
        source,
        pattern,
        find_in_1keys_2values=1,
        nested=False,
        get_flatten=False,
        return_first_found=False,
        return_first_found_path_list=False
):       # starichenko
    """из исходного словаря выбрать только те ключи которые удовлетворяют FullMatch паттерну.

    :param find_in_1keys_2values:
    :param nested: если нужно проверить все вложенные словари - собственно без этого нет особой необходимости
    :param get_flatten: если хотим все привести к одноуровневому словарю

    :param pattern:
        пример r".*\[(.+?)(#[^\]]+)?\].*"
        простой пример r".*_hello_.*"
    """
    result_dict = {}

    for key, value in source.items():
        if nested and isinstance(value, dict):
            result_down_level = dict_find_pattern(
                source=value,
                pattern=pattern,
                find_in_1keys_2values=find_in_1keys_2values,
                nested=nested,
                return_first_found=return_first_found,
                return_first_found_path_list=return_first_found_path_list)

            if return_first_found and result_down_level:
                return {key: result_down_level}
            elif return_first_found_path_list and result_down_level:
                return [key, *result_down_level]
            else:
                result_dict[key] = result_down_level
        else:
            key_str = str(key)
            value_str = str(value)

            if find_in_1keys_2values == 1:
                source_str = value_str
            else:
                source_str = key_str

            match = re.fullmatch(pattern, source_str)
            if match:
                if return_first_found:
                    return value
                elif return_first_found_path_list:
                    return [key, ]
                else:
                    print(f"найден подходящий ключ=[{key}] со значением=[{value}] идем дальше")
                    if key not in result_dict or value == result_dict[key]:
                        result_dict.update({key: value})

    # использовать только в конце!!! чтобы можно было использовать словари с одинаковыми ключами в разных глубинах и разными значениями!
    if get_flatten:
        result_dict = dict_flatten(result_dict)

    return result_dict


# dict-SORT ORDER
def dict_order_by_key_list(
        source,
        keys_list=None,
        raise_if_not_exists=True,
        delete_keys_not_listed=False
):        # starichenko
    """reorder dict by keys_list
    typically transform keys in other func and use this funс to order original dict keys!!!"""
    result_dict = dict()
    for key in keys_list:
        if key in source:
            result_dict.update({key: source.pop(key)})
        elif raise_if_not_exists:
            raise Exception(f"not exists key={key}")
    if not delete_keys_not_listed:
        result_dict.update(source)
    return result_dict


def dict_sort_by_num_keys(source):        # starichenko
    """"""
    sorted_keys_list = sorted(source)
    sorted_dict = dict_order_by_key_list(source, keys_list=sorted_keys_list)
    return sorted_dict


# dict-KEYS
def dict_keys_delete_by_name(source, keys_list_or_func_link=None, reverse_decision=False, nested=True):  # starichenko
    """удалить ключи из словаря.

    применяется когда нужно очистить словарь от лишних данных,
    например если мы хотим сделать Flatten словарь в котором есть Title - то там будут ошибки наличия ключей с разными значениями!!

    :param source: source
    :param keys_list_or_func_link: list of keys to delete? to delete None - use [None]!
    :param reverse_decision: if True - delete all keys except listed in keys_list/funk link
    :param nested: if False - walk only root level of source
    """
    # INPUT -----------------------------------------------------------------------------------------------------------
    if not isinstance(source, dict):
        return source

    if not keys_list_or_func_link:
        return source

    result_dict = copy.deepcopy(source)

    # WORK ------------------------------------------------------------------------------------------------------------
    for key in source:
        if keys_list_or_func_link:
            if callable(keys_list_or_func_link):
                decide = keys_list_or_func_link(key)
            elif isinstance(keys_list_or_func_link, (list, tuple, set)):
                decide = key in keys_list_or_func_link
            else:
                decide = key == keys_list_or_func_link
        else:
            decide = False

        if reverse_decision:
            decide = not decide

        if decide:
            result_dict.pop(key)
            continue

    # NESTED ----------------------------------------------------------------------------------------------------------
    if nested:
        for key, value in result_dict.items():
            if isinstance(value, dict):
                result_dict[key] = dict_keys_delete_by_name(
                    source=value,
                    keys_list_or_func_link=keys_list_or_func_link,
                    reverse_decision=reverse_decision,
                    nested=nested
                )

    print(f"{result_dict=}")
    return result_dict


def dict_keys_delete_by_values(**kwargs):        # starichenko
    # raise Exception("use directly dict_values_replace with param DeleteKeys=True")
    return dict_values_replace(**kwargs, delete_found_keys=True)


def dict_key_rename__by_name(source: dict, key: Any, new_key: Any) -> dict:  # starichenko
    result = {}
    for _key, value in source.items():
        if key == _key:
            _key = new_key
        result.update({_key: value})
    return result


def dict_keys_rename__by_func(source: dict, func: Callable[[Any], Any]) -> dict:  # starichenko
    result = {}
    for key, value in source.items():
        key = func(key)
        result.update({key: value})
    return result


def dict_key_get_short_name(key_fullname, subkeys_list=[]):  # starichenko
    """return short name for source key name
    """
    short_name_pattern = r".*\[(.+?)(#[^\]]+)?\].*"

    if isinstance(subkeys_list, (list, dict, tuple, set)):
        subkeys_list = [str(i) for i in subkeys_list]
    else:
        raise Exception(f"неверное значение параметров subkeys_list={subkeys_list}")

    name_short = None

    match = re.fullmatch(short_name_pattern, key_fullname)
    if match:
        name_short_temp = match.group(1)
        subkeys_string_actual = match.group(2)
        if subkeys_string_actual:
            subkeys_list_actual = sequence_delete_items_blank(subkeys_string_actual.replace(' ', '').split("#"))
            print(f"-требуется совпадение ключей subkeys_list_actual={subkeys_list_actual}")

            if sequence_check_included_in_supersequence(subkeys_list_actual, subkeys_list):
                print(
                    f"+все ключи {subkeys_list_actual} имеются в subkeys_list={subkeys_list} key_fullname=[{key_fullname}]")
                name_short = name_short_temp
            else:
                print(
                    f"-НЕ ДОСТАТОЧНО исходных ключей subkeys_list={subkeys_list} в key_fullname=[{key_fullname}] идем дальше")
        else:
            print(f"+НЕ требуется совпадения ключей")
            name_short = name_short_temp

    return name_short


def dict_keys_short_names(source, subkeys_list=[], get_flatten=True):   # starichenko
    """выдает словарь с короткими именами ключей (именованными параметрами) если они имеются.
    если параметр не имеет короткого имени - ключ игнорируется!
    все настройки имеющие ключи, если имеют хотябы один ключ не из списка subkeys_list - игнорируются!

    если значения одинаковы - спокойно пропускаем!

    :param source: исходный словарь
    :param subkeys_list: дополнительные параемтры для фильтра,
        фильтр не обязывающий! а позволяющий.
        суть фильтра - если исходный ключ имеет какието подключи, то ключ будет принят, только если ВСЕ они находятся в списке разрешенных!
    :param get_flatten: выдать результат в виде словаря одного уровня

    Пример:
        из словаря  = {"[ps_power_load_psi#1200] Требуемая нагрузка, Вт": 900}
        для subkeys_list=[1200, DC] => {"ps_power_load_psi": 900}
        для subkeys_list=[1600, DC] => {}
    """
    result_dict = dict()

    if get_flatten:
        source = dict_flatten(source)

    for key_fullname, value in source.items():
        key_fullname = str(key_fullname)
        print(f"смотрим [{key_fullname}]")

        key_short_name = dict_key_get_short_name(key_fullname, subkeys_list)
        if key_short_name is not None:
            if key_short_name not in result_dict:
                print(f"+найдено первое значение [{key_fullname}] идем дальше")
                result_dict.update({key_short_name: value})
            elif result_dict[key_short_name] == value:
                pass    # same value - pass silent!
            else:
                print(f"result_dict+key_short_name=[{key_short_name}]=", result_dict[key_short_name])
                print(f"source+key_fullname=[{key_fullname}]=", value)
                logging_and_print_warning(f"-в словаре имеется НЕСКОЛЬКО СООТВЕТСТВУЮЩИХ нам параметров с именем {key_short_name} - выходим")
                raise Exception(f"-в словаре имеется НЕСКОЛЬКО СООТВЕТСТВУЮЩИХ нам параметров с именем {key_short_name} - выходим")
        else:
            print(f"в словаре найден элемент без именованного значения (неподходящего для subkeys_list={subkeys_list})! [{key_fullname}]")

    if result_dict == dict():
        logging_and_print_warning(f"---в словаре НЕ НАЙДЕНО именованных параметров с ключами subkeys_list={subkeys_list}")
    else:
        print(f"+в словаре НАЙДЕНЫ ИМЕНОВАННЫЕ параметры с subkeys_list={subkeys_list} result_dict={result_dict}")

    return result_dict


# COMPARE ----------------------------
def dicts_compare_full(
        target: dict,
        compare: dict,
        func: Optional[Callable[[Any, Any], bool]] = None,
) -> "Result":    # starichenko
    """Compare dicts 1 level.

    :param target:
    :param compare:
    :param func: used to compare values
    """
    class Result:
        equel_by_values: dict = {}
        diff_by_values: dict = {}
        intersection: dict = {}     # updated intersection!
        not_in_a: dict = {}
        not_in_b: dict = {}

    result = Result()

    if func is None:
        func = lambda x1, x2: x1 == x2

    for key, value in target.items():
        if key not in compare:
            result.not_in_b.update({key: value})
            continue

        result.intersection.update({key: value})

        if func(value, compare.get(key)):
            result.equel_by_values.update({key: value})
        else:
            result.diff_by_values.update({key: value})

    # NOT_IN_A --------------------------------------------
    for key, value in compare.items():
        if key not in target:
            result.not_in_a.update({key: value})

    return result


def dict_compare_by_same_values(
        source: dict,
        compare: dict,
        func: Optional[Callable[[Any, Any], bool]] = None,
        return_by: int = 0
) -> dict:    # starichenko
    """Return new dict by Delete from source dict keys with same values from compare dict!

    specially created for finding params not already set on Volga!

    :param source:
    :param compare:
    :param func: used to compare values
    :param return_by:
        - 0: all
        - 1: same
        - 2: diff
    :return: new dict
    """
    result = {}

    # if func is None:
    #     func = lambda x1, x2: x1 == x2
    #
    # for key, value in source.items():
    #     if func(value, compare.get(key)):
    #         result.update({key: value})
    return result


def dicts_compare(dict1, dict2, use_only_keys_list=None):     # starichenko
    """
    if dicts eaual - then fully equivalent of direct comparing!
    but if not - will show the different keys and different values!

    :param dict1:
    :param dict2:
    :param keys_list:  if specified compare only this keys!
    :return:
    """
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        msg = f"incorrect input types {dict1=}/{dict2=}"
        logging_and_print_warning(msg)
        return

    dict1 = copy.deepcopy(dict1)
    dict2 = copy.deepcopy(dict2)

    if use_only_keys_list:
        dict1 = dict_keys_delete_by_name(dict1, keys_list_or_func_link=use_only_keys_list, reverse_decision=True)
        dict2 = dict_keys_delete_by_name(dict2, keys_list_or_func_link=use_only_keys_list, reverse_decision=True)

    if dict1 == dict2:
        return True

    # diff_keys -------------------------------------------------------------------------------------------------------
    diff_keys = sequences_get_different_elements(dict1, dict2)

    # diff_values -----------------------------------------------------------------------------------------------------
    diff_values = {}
    for key, value in dict1.items():
        if key not in diff_keys:
            if value != dict2.get(key):
                diff_values.update({key: [value, dict2.get(key)]})
                msg = f"incorrect value dict1({key}:{value}) != dict2({key}:{dict2.get(key)})"
                logging_and_print_warning(msg)

    if diff_keys or diff_values:
        return False


def dict_validate_by_etalon(source, etalon, allow_extra_keys=True, nested=True):  #starichenko
    """

    KEYS compare for existance!
    VALUES compare for types
        VALUE is correct always if any of value_* is None otherwise it must be equal types!!!

    if values in etalon_dict is dict - validation will continue by recursion!

    specially created for validating dicts in testcase

    :param source:
    :param etalon: values must be exact types or None for any type!
    :return:
    """
    result = True

    if not isinstance(source, dict) or not isinstance(etalon, dict):
        msg = f"incorrect input types {source=}/{etalon=}"
        logging_and_print_warning(msg)
        return False

    for key_etalon, value_etalon in etalon.items():
        if key_etalon not in source:
            msg = f"missing expected {key_etalon=}"
            logging_and_print_warning(msg)
            result = False
        else:

            value_source = source.get(key_etalon)
            if None not in [value_source, value_etalon]:
                if type(value_source) != type(value_etalon):
                    msg = f"different types {value_source=}/{value_etalon=}"
                    logging_and_print_warning(msg)
                    result = False
                elif nested:
                    if isinstance(value_etalon, dict):
                        result &= dict_validate_by_etalon(source=value_source, etalon=value_etalon, allow_extra_keys=allow_extra_keys, nested=nested)
                    elif isinstance(value_etalon, (set, list)) and len(value_etalon) == 1 and isinstance(value_etalon[0], dict):
                        result &= dicts_list_validate_by_etalon(source=value_source, etalon=value_etalon[0], allow_extra_keys=allow_extra_keys, nested=nested)

    # allow_extra_keys
    if not allow_extra_keys:
        for key_source in source:
            if key_source not in etalon:
                msg = f"not allowed {key_source=}"
                logging_and_print_warning(msg)
                result = False

    return result


def dicts_list_validate_by_etalon(source, etalon, allow_extra_keys=True, nested=True):  #starichenko
    if source is None:
        return True
    if not isinstance(source, (set, list)):
        return

    if len(source) == 0:
        return True

    result = True
    for item in source:
        result &= dict_validate_by_etalon(
            source=item,
            etalon=etalon,
            allow_extra_keys=allow_extra_keys,
            nested=nested)

    return result


# dict-MERGE+UPDATES --------------------
def dicts_merge(source_list):     # starichenko
    """
    merge dicts

    1. if any dict have valuetype=dict - other dicts MUST have:
        1. same type - will launch recursion
        2. or any blank value - will assumed as blank dict and recursion applied
        3. in other types - process will be stopped and return None
    2. if no one dict in source_list have dict type in considered value all values will be REPLACED!
        no matter was it list and int or both lists!
    """
    for item_dict in source_list:
        if not isinstance(item_dict, (dict,)):
            msg = f"incorrect types {item_dict=}"
            logging_and_print_warning(msg)
            return

    result_dict = {}
    for new_dict in source_list:
        for new_key, new_value in new_dict.items():
            if new_key not in result_dict:
                result_dict[new_key] = new_value
            elif isinstance(result_dict.get(new_key), dict) and isinstance(new_value, dict):
                # recursion
                result_dict[new_key] = dicts_merge(source_list=[result_dict.get(new_key), new_value])
            else:
                result_dict[new_key] = new_value

    return result_dict


# dict-MAKE
def dict_make_simple_example():     # Starichenko
    sample_dict = {'param1': 'Value1str', 'param2': 123, 'param3': {1: 1, 2: 2}, 'param4': ['list_value1', 'list_value2']}
    return sample_dict


def dict_make_from_keys_list_with_none_values(keys_list):
    """ Returns dict of keys with empty values.
    """
    result_dict = {}
    for key in keys_list:
        result_dict[key] = None
    return result_dict


def dict_ensure_first_key(source, first_key=None):
    """ Returns dict with first key as 'first_key' if present.
    """
    result_dict = dict()

    if first_key and first_key in source:
        result_dict[first_key] = source[first_key]

    for key, value in source.items():
        result_dict[key] = value
    return result_dict


# dict-VALUES
def dict_value_get_with_default_and_replacing(source, key, default=None, replace_list=[], replace_dict={}):    # starichenko
    """
    useful if its not enough for you standard GET method which give you default only if key is not exist in dict

    :param source: source dict
    :param key: key to get value
    :param default: the same default param as for GET method
    :param replace_list: sequence or results which you want to replace (useful for None result)
    :param replace_dict: dict value-result you want to replace in final value from source
    """
    type_check_by_variants(source, (dict,))

    # 1=step default
    result = source.get(key, default)

    # 2=step replace_list
    if result in replace_list:
        result = default

    # 2=step replace_dict
    result = value_try_replace_by_dict(source=result, search_dict=replace_dict, search_type_1starts_2full_3ends_4any_5fullmatch=2)
    return result


def dict_value_get_by_keypath(source: dict, keypath: Union[str, list]) -> Optional[ny]:      # Starichenko
    """get value from dictionary specifying by keys like a path

    example:
        for {1:1, "2":{3:3}}
        with keypath=["2", 3]
        result = 3

    CAUTION: RETURN DICTS DIRECT ACCESS!!!
    """
    # INPUT ---------------------------------------------------
    value_temp = source
    keypath = sequence_make_ensured_if_not(keypath)
    if not keypath:
        msg = f"blank {keypath=}"
        logging_and_print_warning(msg)
        return

    # WORK ---------------------------------------------------
    try:
        for key in keypath:
            value_temp = value_temp[key]
    except Exception as exc:
        msg = f"no {keypath=} in {source=}\n{exc!r}"
        logging_and_print_warning(msg)
        return

    # RESULT ---------------------------------------------------
    return value_temp


def dict_values_check_all_none(dictionary, nested=False):
    """ Returns True if all values of given dictionary are None
        and False otherwise.

    :param nested:
        if False - walk only 1 level and nested dicts are not acceptable! because will return False!
        if True - nested dict is acceptable! will walk all nested dicts
    """
    for key, value in dictionary.items():
        if nested and isinstance(value, dict):
            recursion_result = dict_values_check_all_none(value, nested=True)
            if not recursion_result:
                return False
        elif value is not None:
            return False
    return True


def dict_values_replace(
        source,
        old_values_list_or_func_link=None,
        new_value_or_func_link=None,
        protect_keys_or_func_link=None,
        nested=True,
        delete_found_keys=None,    # dont deprecate! its necessary!
):     # Starichenko
    """walk dict and rewrite exact values by exact new value
    If you need replace ALL VALUES use old_values_list_or_func_link = lambdaTrue!

    :param old_values_list_or_func_link: if value is in here - rewrite value!
    :param protect_keys_or_func_link: dont touch values if so!
    :param delete_found_keys: if value Matched - key will be deleted
    """
    if not isinstance(source, dict):
        return source

    result_dict = copy.deepcopy(source)

    for key, value in source.items():
        # decide ---------------------------
        # value
        if callable(old_values_list_or_func_link):
            decision_v = old_values_list_or_func_link(value)
        elif isinstance(old_values_list_or_func_link, (list, tuple, set)):
            decision_v = value in old_values_list_or_func_link
        else:
            decision_v = value == old_values_list_or_func_link

        # key
        if protect_keys_or_func_link:
            if callable(protect_keys_or_func_link):
                protect_k = protect_keys_or_func_link(key)
            elif isinstance(protect_keys_or_func_link, (list, tuple, set)):
                protect_k = key in protect_keys_or_func_link
            else:
                protect_k = key == protect_keys_or_func_link
        else:
            protect_k = False

        # work -----------------------------
        if decision_v and not protect_k:
            if delete_found_keys:
                result_dict.pop(key)
            else:
                if callable(new_value_or_func_link):
                    result_dict[key] = new_value_or_func_link(value)
                else:
                    result_dict[key] = new_value_or_func_link
            continue

        # nested ---------------------------
        if isinstance(value, dict) and nested:
            result_dict[key] = dict_values_replace(
                source=value,
                old_values_list_or_func_link=old_values_list_or_func_link,
                new_value_or_func_link=new_value_or_func_link,
                protect_keys_or_func_link=protect_keys_or_func_link,
                nested=nested,
                delete_found_keys=delete_found_keys,
            )

    return result_dict


def dict_values_get_list_for_key(
        source,
        key,
        return_first_value=False,
        no_repetitions=True,
        use_blank=False):             # Starichenko
    """выдает списоком значения всех ключей с именем key в многооуровневом словаре.
    результат в списке. если ничего не найдено? то выдаст пустой список [].

    :param return_first_value:
        для случаев, когда подразумевается, что в словаре в разных местах (в разной глубине) вставлены одни и теже значения для одного ключа!
    :param no_repetitions: если выводится несколько значений, то если уже есть такое значение, то не добавляется!
        можно использовать повторяющиеся значения для дальнейшего анализа их количенства
    """
    key_find = key
    value_list = []
    if not isinstance(source, dict):
        return

    # level root
    if key_find in source:
        value = source[key_find]
        if not value_is_blanked(value, zero_as_blank=False) or use_blank:
            if no_repetitions and value in value_list:
                pass
            else:
                value_list.append(value)
            if return_first_value:
                return value

    # level nested
    for key, value in source.items():
        if isinstance(value, dict):
            subresult_list = dict_values_get_list_for_key(
                source=value,
                key=key_find,
                return_first_value=return_first_value,
                no_repetitions=no_repetitions,
                use_blank=use_blank
            )
            if return_first_value and subresult_list:
                return subresult_list
            else:
                value_list = lists_sum([value_list, subresult_list], no_repetitions=no_repetitions)

    print(f"source[{key_find}]list={value_list=}")
    return value_list


def dict_value_get_by_short_key(name, source,
                                 dict_already_short=False,
                                 subkeys_list=[],
                                 nested=True,
                                 return_first_value=True,
                                 no_repetitions=True):  # STARICHENKO
    """функция ищет указанное имя с обрамляющими квадратными скобками в Settings.
    И выдает его значение.

    :raw_dict/short_dict: используется только одно значение!!!
        если raw_dict - то из него формируется short_dict

    subkeys_list: список ключей который могут быть полезны для определения значения (не все могут быть использованы!)
        для передачи этих ключей рекомендуется создавать отдельный параметр в самом DUT
    :nested: если False - смотрит только корень переданного словаря! если True - смотрит еще и все вложенные словари!
    :dont_give_error_if_found_same_name_with_same_value: вообще лучше если будет предупреждение по поводу одинакового имени в параметрах,
        но скорее всего это будет часто происходить но с одинаковыми значениями, и лучше его отключить по умолчанию
        при разных значениях ошибка в любом случае Критическая и обязана быть!

    отличия от get_value_named_setting_with_keys:
    1. работает независимо от количества ключей
    2. при наличии любого количества ключей в рассматриваемой строке Settings - смотрится есть ли они в списке subkeys_list
    3. если все ключи есть - параметр принимается, если нет - анализ останавливается и переходим к следующей строке
    4. если втречается дважды полностью подходящее значение - вызывается ошибка!

    ВЫВОД:
    ключи должны быть уникальными, тоесть однозначно должно быть понятно что они подразумевают без контекста!
    тоесть "DC" "1200" - всем понятные контексты!
    """
    name = str(name)

    # todo: dont short dict, just only find exact names!!!
    if dict_already_short:
        short_dict = source
    else:
        if nested:
            raw_dict = dict_flatten(source)
        short_dict = dict_keys_short_names(source=raw_dict, subkeys_list=subkeys_list)

    if name in short_dict:
        return short_dict[name]
    else:
        logging_and_print_warning(f"нет в словаре нужной переменной [{name}]")
        return


# dict-CONVERTION
def dict_unique__collapse_flatten_by_level(source: dict, level: int = -1):    #starichenko
    """
    works in dicts with UNIQUE keys by max level deep.

    special created for getting Volga params info from sections GDDU

    :param source:
    :param level: use negative for all levels!
    :return:
    """
    result = dict()

    if level == 0:
        return source

    for key, value in source.items():
        if not isinstance(value, dict):
            result[key] = value
            continue

        value = dict_unique__collapse_flatten_by_level(source=value, level=level-1)
        result.update(value)
    return result


def dict_flatten(source,
                 keys_list_del=[],
                 keys_list_update=[],
                 keys_list_dont_go_deep=[],
                 new_values_update=False,
                 new_values_replace_by="#ОШИБКА#",
                 _is_recursion_applied=False,
                 _result_dict={}):                  # Starichenko
    """
    преобразует все сложенные словари в словарь одного уровня вложенности!
    т.е. все ключи из всех словарей выводит в один словарь.

    :keys_list_del: прежде чем чтото будет делаться - очищается весь словарь от этих ключей!
    :keys_list_update: все ключи из этого списка будут перезаписаны в результирующем словаре, если будут иметь разные значения, независимо от других настроек функции
    :keys_list_dont_go_deep: if this keys have dict-value don't go deep, use like ordinary value
    :_is_recursion_applied: mark function as recursion step! dont use from outside!
    :_result_dict: for marking when recurtion is started!
    """
    source = dict_keys_delete_by_name(source=source, keys_list_or_func_link=keys_list_del)

    result_dict = _result_dict
    for key, value in source.items():
        if isinstance(value, dict) and key not in keys_list_dont_go_deep:
            result_dict = dict_flatten(source=value,
                                       keys_list_del=keys_list_del,
                                       keys_list_update=keys_list_update,
                                       keys_list_dont_go_deep=keys_list_dont_go_deep,
                                       new_values_update=new_values_update,
                                       new_values_replace_by=new_values_replace_by,
                                       _is_recursion_applied=True,
                                       _result_dict=result_dict)
        else:
            if key not in result_dict or key in keys_list_update:
                result_dict.update({key: value})
            elif result_dict[key] == source[key]:
                pass
            else:       # если значения не сходятся!
                logging_and_print_warning("error!!! detected same keys with DIFFERENT values")
                logging_and_print_warning(dict_get_pretty_string(result_dict))
                msg = F"1, key=[{key}] old_value result_dict[key]=[{result_dict[key]}]"
                logging_and_print_warning(msg)
                msg = F"2, key=[{key}] new_value=[{value}]"
                logging_and_print_warning(msg)
                if new_values_update:
                    result_dict.update({key: value})
                elif new_values_replace_by:
                    result_dict.update({key: new_values_replace_by})
                    msg = F"3, key=[{key}] OVERRITEN_value=[{new_values_replace_by}]"
                    logging_and_print_warning(msg)

    if not _is_recursion_applied:
        print(f"result_dict={result_dict}")
    return result_dict


def dict_get_pretty_string(source, name=None, _print=False):   # starichenko
    """return pretty string for dict
    """
    try:
        pretty_string = json.dumps(source, indent=4)
    except:
        msg = f"CANT jsonDUMP source=[{source}]"
        logging_and_print_warning(msg)
        pretty_string = str(source)

    if name:
        pretty_string = f"{name} = {pretty_string}"

    if _print:
        print(pretty_string)
    return pretty_string


# dict-COMBINATIONS
def dict_iter_combinations_by_key_list_and_value_list(source=None):   # starichenko
    """return all combinations between key+value lists.
    can be used for brute force login+password.

    Example:
        {1: 2, 10: [11, 12], (21, 22): 23, (31, 32): (33, 34)}
        [(1, 2),    (10, 11), (10, 12),    (21, 23), (22, 23),     (31, 33), (31, 34), (32, 33), (32, 34),]
    """
    if source is None:
        source = source = {1: 2, 10: [11, 12], (21, 22): 23, (31, 32): (33, 34)}

    for key_seq, value_seq in source.items():

        if not type_is_iterable_but_not_str(key_seq):
            key_seq = [key_seq, ]
        if not type_is_iterable_but_not_str(value_seq):
            value_seq = [value_seq, ]

        for key in key_seq:
            for value in value_seq:
                yield (key, value)


def dict_make_string_of_values_if_object(source):  # starichenko
    """
    if you want to dump dict wich can have values as object like pathlib - you will use it!

    :param source:
    :return:
    """
    result = dict_values_replace(
        source=source,
        old_values_list_or_func_link=lambda _value: not type_is_elementary_single_or_container(_value),
        new_value_or_func_link=str
    )
    return result


# =====================================================================================================================



from base_aux.aux_argskwargs.m4_kwargs_eq_expect import *

from base_aux.cli.m1_cli_user import *


# =====================================================================================================================
if KwargsEqExpect_OS().bool_if__any_true(windows=True):
    CMD_PING_1 = "ping -n 1 localhost"      #momentary! less then 0.1sec!
    CMD_PING_2 = "ping -n 2 localhost"      #about 1sec!
else:
    CMD_PING_1 = "ping -c 1 localhost"
    CMD_PING_2 = "ping -c 2 localhost"


# =====================================================================================================================
class Test:
    def test__ok(self):
        victim = CliUser()

        assert victim.send(CMD_PING_1, timeout=2)
        assert victim.last_cmd == CMD_PING_1
        assert victim.last_finished is True

        assert victim.last_exc_timeout is None
        assert bool(victim.last_stdout) is True
        assert bool(victim.last_stderr) is False
        assert victim.last_retcode == 0
        assert victim.last_finished_success
        assert victim.counter == 1
        assert victim.counter_in_list == 0

    def test__count(self):
        victim = CliUser()
        assert victim.counter == 0
        assert victim.counter_in_list == 0

        assert victim.send(CMD_PING_1, timeout=1)
        assert victim.counter == 1
        assert victim.counter_in_list == 0

        assert victim.send(CMD_PING_1, timeout=1)
        assert victim.counter == 2
        assert victim.counter_in_list == 0

        assert victim.send([CMD_PING_1, CMD_PING_1], timeout=1)
        assert victim.counter == 4
        assert victim.counter_in_list == 2

        assert victim.send(CMD_PING_1, timeout=1)
        assert victim.counter == 5
        assert victim.counter_in_list == 0

        assert victim.send([CMD_PING_1, CMD_PING_1, CMD_PING_1], timeout=1)
        assert victim.counter == 8
        assert victim.counter_in_list == 3

        assert not victim.send([CMD_PING_2, CMD_PING_2, CMD_PING_2], timeout=1)
        assert victim.counter == 9
        assert victim.counter_in_list == 1

    def test__list(self):
        victim = CliUser()

        assert not victim.send([CMD_PING_1, CMD_PING_2], timeout=1)
        assert victim.counter_in_list == 2

        assert not victim.send([CMD_PING_1, CMD_PING_2, CMD_PING_2], timeout=2)
        assert victim.counter_in_list == 3

        assert victim.send([CMD_PING_1, CMD_PING_2], timeout=2)
        assert victim.counter_in_list == 2
        assert victim.last_cmd == CMD_PING_2
        assert victim.last_finished is True

        assert victim.last_exc_timeout is None
        assert bool(victim.last_stdout) is True
        assert bool(victim.last_stderr) is False
        assert victim.last_retcode == 0
        assert victim.last_finished_success

    def test__list_not_passed_timeout(self):
        victim = CliUser()
        assert not victim.send([CMD_PING_1, CMD_PING_2], timeout=0.1)
        assert victim.send([CMD_PING_1, CMD_PING_2])

    @pytest.mark.parametrize(
        argnames="cmds, timeout_def, _EXPECTED",
        argvalues=[
            # FIXME:HERE - NOT ALWAYS PASSED!!! dont panic! maybe need to skip it or ref!
            # FIXME:HERE - NOT ALWAYS PASSED!!! dont panic! maybe need to skip it or ref!
            # FIXME:HERE - NOT ALWAYS PASSED!!! dont panic! maybe need to skip it or ref!
            # FIXME:HERE - NOT ALWAYS PASSED!!! dont panic! maybe need to skip it or ref!
            # FIXME:HERE - NOT ALWAYS PASSED!!! dont panic! maybe need to skip it or ref!
            # FIXME:HERE - NOT ALWAYS PASSED!!! dont panic! maybe need to skip it or ref!
            # FIXME:HERE - NOT ALWAYS PASSED!!! dont panic! maybe need to skip it or ref!
            (CMD_PING_2, 0.1, False),
            ((CMD_PING_2, 0.1), 0.1, False),
            ((CMD_PING_2, 2), 0.1, True),

            # ([(CMD_PING_1, 0.1), CMD_PING_1], 0.1, True),      # here is wrong/now i dont anderstand what is goingOn/testing here!??????
            ([(CMD_PING_1, 0.1), CMD_PING_2], 0.1, False),
            ([(CMD_PING_1, 0.1), (CMD_PING_2, 2)], 0.1, True),
            ([(CMD_PING_1, 0.1), (CMD_PING_2, None)], 0.1, False),
            ([(CMD_PING_1, 0.1), (CMD_PING_2, None)], 2, True),
        ]
    )
    def test__tuple(self, cmds, timeout_def, _EXPECTED):
        func_link = CliUser().send(cmd=cmds, timeout=timeout_def)
        Lambda(func_link).check_expected__assert(_EXPECTED)

    def test__list__till_first_true(self):
        victim = CliUser()

        assert not victim.send([CMD_PING_1, CMD_PING_2], timeout=1)
        assert victim.counter_in_list == 2

        assert victim.send([CMD_PING_1, CMD_PING_2], timeout=1, till_first_true=True)
        assert victim.counter_in_list == 1

        assert victim.last_cmd == CMD_PING_1
        assert victim.last_finished is True

        assert victim.last_exc_timeout is None
        assert bool(victim.last_stdout) is True
        assert bool(victim.last_stderr) is False
        assert victim.last_retcode == 0
        assert victim.last_finished_success

    def test__exc_timeout(self):
        victim = CliUser()

        assert not victim.send(CMD_PING_2, timeout=0.1)
        assert victim.last_cmd == CMD_PING_2
        assert victim.last_finished is True

        assert isinstance(victim.last_exc_timeout, Exc_CliTimeout)
        # assert bool(victim.last_stdout) is False
        assert bool(victim.last_stderr) is False
        assert victim.last_retcode is None
        assert not victim.last_finished_success

    def test__exc_not_exists(self):
        victim = CliUser()

        cmd_line = "ping123"
        assert not victim.send(cmd_line, timeout=10)
        assert victim.last_cmd == cmd_line
        assert victim.last_finished is True

        assert victim.last_exc_timeout is None
        # assert bool(victim.last_stdout) is False
        assert bool(victim.last_stderr) is True
        assert victim.last_retcode not in [0, None]
        assert not victim.last_finished_success

    def test__exc_cli_available(self):
        # one cmd ------------------------------------------------
        class CliUserForAvailable(CliUser):
            CMDS_REQUIRED = {"ping123": None, }

        try:
            victim = CliUserForAvailable()
        except Exception as exc:
            assert isinstance(exc, Exc_CliNotAvailable)

        # two cmd ------------------------------------------------
        class CliUserForAvailable(CliUser):
            CMDS_REQUIRED = {CMD_PING_1: None, }

        victim = CliUserForAvailable()
        assert victim.cli_check_available()


# =====================================================================================================================



from base_aux.base_values.m3_exceptions import *
from base_aux.base_nest_dunders.m1_init0_annots_required import *


# =====================================================================================================================
def test__raise():
    class Victim(NestInit_AnnotsRequired):
        ATTR1: int
        ATTR2: int = 2

    try:
        Victim()
    except Exc__NotExistsNotFoundNotCreated:
        assert True
    else:
        assert False


def test__ok():
    class Victim(NestInit_AnnotsRequired):
        ATTR1: int = 1
        ATTR2: int = 2

    try:
        Victim()
    except Exc__NotExistsNotFoundNotCreated:
        assert False
    except:
        pass


def test__NT():
    try:
        class Victim(NestInit_AnnotsRequired, NamedTuple):
            ATTR1: int
            ATTR2: int = 2
        assert False
    except TypeError:
        # TypeError: can only inherit from a NamedTuple type and Generic
        pass


# =====================================================================================================================
# @pytest.mark.skip
# class Test:
#     # FIXME: make a ref!
#     # -----------------------------------------------------------------------------------------------------------------
#     def test__dataclass(self):
#         @dataclass
#         class Cls(NestInit_AnnotsRequired):
#             ATTR1: int
#             ATTR2: int = 2
#
#         assert Cls(1).annots_get_set() == {"ATTR1", }
#         assert Cls(1).annots_get_dict() == {"ATTR1": 1, }
#
#     def test__dataclass_by_obj(self):
#         # 10------------------
#         @dataclass
#         class Cls:
#             ATTR1: int
#             ATTR2: int = 2
#
#         class Cls2(Cls):
#             ATTR2: int = 22
#             ATTR3: int = 33
#
#         obj = Cls2(1)
#         assert NestInit_AnnotsRequired().annots_get_set(obj) == {"ATTR1", }
#         assert NestInit_AnnotsRequired().annots_get_dict(obj) == {"ATTR1": 1, }
#
#         # 01------------------
#         class Cls:
#             ATTR1: int
#             ATTR2: int = 2
#
#         @dataclass
#         class Cls2(Cls):
#             ATTR2: int = 22
#             ATTR3: int = 33
#
#         obj = Cls2(1)
#         assert NestInit_AnnotsRequired().annots_get_set(obj) == {"ATTR1", }
#         try:
#             assert NestInit_AnnotsRequired().annots_get_dict(obj) == {"ATTR1": 1, }
#         except Exc__NotExistsNotFoundNotCreated:
#             pass   # its GOOD!!!
#         else:
#             assert False
#
#         # 11------------------
#         @dataclass
#         class Cls:
#             ATTR1: int
#             ATTR2: int = 2
#
#         @dataclass
#         class Cls2(Cls):
#             ATTR2: int = 22
#             ATTR3: int = 33
#
#         obj = Cls2(1)
#         assert NestInit_AnnotsRequired().annots_get_set(obj) == {"ATTR1", }
#         assert NestInit_AnnotsRequired().annots_get_dict(obj) == {"ATTR1": 1, }
#
#     def test__PROPERTY_w_ITER_w_VALUES(self):
#         # 1---------------------------------------------------------
#         # this will work correctly
#         class Cls:
#             ATTR1: int
#             ATTR2: int = 2
#
#             @property
#             def meth_as_property(self):
#                 return 333
#
#         obj = Cls()
#         obj.ATTR1 = 1
#
#         assert NestInit_AnnotsRequired().annots_get_set(obj) == {"ATTR1", }
#         assert NestInit_AnnotsRequired().annots_get_dict(obj) == {"ATTR1": 1, }
#         assert list(NestInit_AnnotsRequired().annots_get_values(obj)) == [1, ]
#
#         # 2---------------------------------------------------------
#         # this will RAISE!
#         class Cls:
#             ATTR1: int
#             ATTR2: int = 2
#
#             @property
#             def meth_as_property(self):
#                 return sum(self)
#
#             def __iter__(self):
#                 yield from NestInit_AnnotsRequired().annots_get_values(self)
#
#         obj = Cls()
#         obj.ATTR1 = 1
#
#         assert NestInit_AnnotsRequired().annots_get_set(obj) == {"ATTR1", }
#
#         try:
#             assert NestInit_AnnotsRequired().annots_get_dict(obj) == {"ATTR1": 1, }
#             # assert list(NestInit_AnnotsRequired().annots_get_values(obj)) == [1, ]
#         except RecursionError:
#             pass
#         else:
#             assert False
#
#         # 3 FIXED---------------------------------------------------------
#         # this will work correctly!
#         class Cls:
#             ATTR1: int
#             ATTR2: int = 2
#
#             def meth_NO_property(self):
#                 return sum(self)
#
#             def __iter__(self):
#                 yield from NestInit_AnnotsRequired().annots_get_values(self)
#
#         obj = Cls()
#         obj.ATTR1 = 1
#
#         assert NestInit_AnnotsRequired().annots_get_set(obj) == {"ATTR1", }
#         assert NestInit_AnnotsRequired().annots_get_dict(obj) == {"ATTR1": 1, }
#         assert list(NestInit_AnnotsRequired().annots_get_values(obj)) == [1, ]


# =====================================================================================================================



from typing import *
import time

from base_aux.base_singletons.m1_singleton import *
from base_aux.aux_argskwargs.m1_argskwargs import *
from base_aux.base_lambdas.m2_lambda3_thread import LambdaThread


# =====================================================================================================================
class ThreadsDecorCollector(Singleton_CallMeta):
    """
    TODO: DEPRECATE??? use clearly direct other methods/objects!

    NOTE
    ----
    1/ maybe you dont need use it if you need only one class method - use direct QThread

    GOAL
    ----
    Manager for spawning threads and keep its instances with additional data (result/exc).
    Singleton! do you dont need saving instances!

    USAGE
    -----
    use different managers for different funcs/methods if needed
    use just one decorator to spawn threads from func / methods
    keep all spawned threads in list by LambdaThread base_types

    1. BEST PRACTICE
    Not recommended using it directly, use as simple nested:
        class ThreadsManager1(ThreadsDecorCollector):
            pass

        @ThreadsManager1().decorator__to_thread
        def func(*args, **kwargs):
            pass

    2. Direct usage
    But if you need only one manager - do use directly without nesting.
        @ThreadsDecorCollector().decorator__to_thread
        def func(*args, **kwargs):
            pass

    :param args: NAME for manager instance
    :param thread_items: LambdaThread instances,
    :param MUTEX: mutex for safe collecting threads in this manager, creates in init
    :param counter: counter for collected threads in this manager

    SPECIALLY CREATED FOR
    ---------------------
    stock strategies/monitors
    """
    THREADS: list[LambdaThread]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.THREADS = []

    @classmethod
    @property
    def NAME(cls) -> str:
        """class name for manager
        """
        return cls.__name__

    @property
    def count(self) -> int:
        return len(self.THREADS)

    # =================================================================================================================
    def decorator__to_thread(self, _func) -> Callable:
        """Decorator which start thread from funcs and methods.

        always collect base_types threads in result object! even if nothread! so you can get results from group!

        :param _func: decorated SOURCE
        """
        def _wrapper__spawn_thread(*args, nothread: bool = None, **kwargs) -> Optional[Any]:
            """actual wrapper which spawn thread from decorated SOURCE.

            :param args: args passed into SOURCE/method,
            :param kwargs: kwargs passed into SOURCE/method,
            """
            thread_item = LambdaThread(_func, *args, **kwargs)
            self.THREADS.append(thread_item)
            thread_item.start()

            if nothread:
                thread_item.wait()
                return thread_item.RESULT

        return _wrapper__spawn_thread

    # =================================================================================================================
    def clear(self) -> None:
        """clear collected thread_items.

        useful if you dont need collected items any more after some step. and need to manage new portion.
        """
        self.THREADS.clear()

    def wait_all(self) -> None:
        """wait while all spawned threads finished.
        """
        # wait all started
        if not self.count:
            time.sleep(0.2)

        for _ in range(3):
            for item in self.THREADS:
                item.wait()

            time.sleep(0.1)

    def terminate_all(self) -> None:
        for thread in self.THREADS:
            thread.terminate()

    def check_results_all(self, value: Any = True, func_validate: Callable[[Any], bool] = None) -> bool:
        """check if result values for all threads are equal to the value

        :param value: expected comparing value for all thread results
        :param func_validate:
        """
        for thread in self.THREADS:
            if func_validate is not None:
                if not func_validate(thread.RESULT):
                    return False
            else:
                if thread.RESULT != value:
                    return False
        return True


# =====================================================================================================================
pass
pass
pass
pass
pass

# EXAMPLES ------------------------------------------------------------------------------------------------------------
count = 5
time_start = time.time()


# define victim ------------------
class ThreadDeCollector1(ThreadsDecorCollector):
    pass


class Cls:
    @ThreadDeCollector1().decorator__to_thread
    def func1(self, num):
        time.sleep(1)
        return num * 1000


# spawn ------------------
for i in range(count):
    assert Cls().func1(i) is None

assert ThreadDeCollector1().count == count
ThreadDeCollector1().wait_all()
assert {item.RESULT for item in ThreadDeCollector1().THREADS} == {num * 1000 for num in range(count)}

ThreadDeCollector1().clear()

# spawn ------------------
for i in range(count):
    assert Cls().func1(i) is None


# =====================================================================================================================



from base_aux.aux_eq.m3_eq_valid2_validators import *
from base_aux.aux_eq.m3_eq_valid1_base import *
from base_aux.base_types.m0_static_types import *
from base_aux.base_enums.m2_enum1_adj import *


# =====================================================================================================================
@final
class EqValid_IsinstanceSameinstance(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.IsinstanceSameinstance


# =====================================================================================================================
@final
class EqValid_Contain(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.Contain


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_ContainStrIc(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.ContainStrIc


# =====================================================================================================================
@final
class EqValid_Startswith(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.Startswith


@final
class EqValid_StartswithIc(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.StartswithIc


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_Endswith(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.Endswith


@final
class EqValid_EndswithIc(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.EndswithIc


# =====================================================================================================================
@final
class EqValid_BoolTrue(Base_EqValid):
    VALIDATOR = Validators.BoolTrue


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_Raise(Base_EqValid):
    VALIDATOR = Validators.Raise


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_NotRaise(Base_EqValid):
    VALIDATOR = Validators.NotRaise


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_Exc(Base_EqValid):
    VALIDATOR = Validators.Exc


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_ExcRaise(Base_EqValid):
    VALIDATOR = Validators.ExcRaise


# =====================================================================================================================
# @final    # DONT USE FINAL!!! need next in - CHAIN!!!
class EqValid_EQ(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.CMP_EQ

@final
class EqValid_EQ_StrIc(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.CMP_EQ__StrIc


# @final    # DONT USE FINAL!!! need next in - EqValid_NumParsedSingle_EQ!!!
class EqValid_EQ_NumParsedSingle(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    VALIDATOR = Validators.CMP_EQ__NumParsedSingle


# =====================================================================================================================
@final
class EqValid_LT(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE
    VALIDATOR = Validators.CMP_LT

@final
class EqValid_LE(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE
    VALIDATOR = Validators.CMP_LE

@final
class EqValid_GT(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE
    VALIDATOR = Validators.CMP_GT

@final
class EqValid_GE(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE
    VALIDATOR = Validators.CMP_GE

# --------------------------------
@final
class EqValid_LGTE(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE
    VALIDATOR = Validators.CMP_LGTE

# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_LT_NumParsedSingle(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE
    VALIDATOR = Validators.CMP_LT_NumParsedSingle

@final
class EqValid_LE_NumParsedSingle(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE
    VALIDATOR = Validators.CMP_LE_NumParsedSingle

@final
class EqValid_GT_NumParsedSingle(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE
    VALIDATOR = Validators.CMP_GT_NumParsedSingle

@final
class EqValid_GE_NumParsedSingle(Base_EqValid):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE
    VALIDATOR = Validators.CMP_GE_NumParsedSingle

# --------------------------------
@final
class EqValid_LGTE_NumParsedSingle(Base_EqValid):
    VALIDATOR = Validators.CMP_LGTE_NumParsedSingle

# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_NumParsedSingle_Success(Base_EqValid):
    VALIDATOR = Validators.NumParsedSingle_Success


@final
class EqValid_NumParsedSingle_TypeInt(Base_EqValid):
    VALIDATOR = Validators.NumParsedSingle_TypeInt


@final
class EqValid_NumParsedSingle_TypeFloat(Base_EqValid):
    VALIDATOR = Validators.NumParsedSingle_TypeFloat

@final
class EqValid_NumParsedSingle_EQ(EqValid_EQ_NumParsedSingle):
    """
    just a link to keep name schema!
    """
    pass


# =====================================================================================================================
class EqValid_Regexp(Base_EqValid):
    """
    NOTE
    ----
    for one regexp - simply use this EqValid_Regexp
    for several patterns - use other classes for clarification!
    """
    VALIDATOR = Validators.Regexp
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE

# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_RegexpAllTrue(EqValid_Regexp):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE


@final
class EqValid_RegexpAnyTrue(EqValid_Regexp):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE


@final
class EqValid_RegexpAllFalse(EqValid_Regexp):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_FALSE


@final
class EqValid_RegexpAnyFalse(EqValid_Regexp):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_FALSE


# =====================================================================================================================
@final
class EqValid_AttrsByKwargs(Base_EqValid):
    VALIDATOR = Validators.AttrsByKwargs


# ---------------------------------------------------------------------------------------------------------------------
# @final
# class EqValid_AttrsByObj(Base_EqValid):
#     VALIDATOR = Validators.AttrsByObj
#     ATTR_LEVEL: EnumAdj_AttrScope = EnumAdj_AttrScope.NOT_PRIVATE


@final
class EqValid_AttrsByObjNotPrivate(Base_EqValid):
    VALIDATOR = Validators.AttrsByObj
    ATTR_LEVEL: EnumAdj_AttrScope = EnumAdj_AttrScope.NOT_PRIVATE


@final
class EqValid_AttrsByObjNotHidden(Base_EqValid):
    VALIDATOR = Validators.AttrsByObj
    ATTR_LEVEL: EnumAdj_AttrScope = EnumAdj_AttrScope.NOT_HIDDEN


def _explore():
    class Cls:
        o = 1
        _h = 1
        __p = 1

    source = Cls()
    other = Cls()
    ev = EqValid_AttrsByObjNotPrivate(source)
    print(f"{ev=}")
    print(f"{ev == other}")


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_AnnotsAllExists(Base_EqValid):
    VALIDATOR = Validators.AnnotsAllExists


# =====================================================================================================================
if __name__ == "__main__":
    _explore()


# =====================================================================================================================



import platform

from base_aux.aux_argskwargs.m3_args_bool_raise_if import *
from base_aux.aux_text.m8_str_ic import *
from base_aux.aux_dict.m2_dict_ic import *


# =====================================================================================================================
TYPING__EQ_VALID__FINAL = Base_EqValid
TYPING__EQ_VALID__DRAFT = Any | Base_EqValid


# =====================================================================================================================
class Base_KwargsEqExpect:
    """
    GOAL
    ----
    1/ use object as named collection of EqValids
    2/ select exact EqValids for final cmp
    3/ select exact results for any expectatins

    SPECIALLY CREATED FOR
    ---------------------
    replace inconvenient Base_ReqCheckStr for objects like ReqCheckStr_Os
    good check exact expectations

    PARAMS
    ------
    EQ_VALID__CLS_DEF
        if used - values in EQ_KWARGS used as arg for it!
        EqValid_EQ_StrIc i  most useful
    """
    OTHER_DRAFT: Any | Callable = True
    OTHER_FINAL__RESOLVE: bool = True
    OTHER_RAISED: bool = None
    OTHER_FINAL: Any | Exception = None

    EQ_VALID__CLS_DEF: type[Base_EqValid] | None = None
    EQ_KWARGS: DictIcKeys[str, TYPING__EQ_VALID__FINAL] = {}      # dont use StrIc! cant direct access as GI(value)
    EQ_EXPECTS: dict[str, bool | None | Any] = {}

    def __init__(self, other_draft: Any | Callable = NoValue, _eq_valid__cls_def: type[Base_EqValid] = None, **eq_kwargs: TYPING__EQ_VALID__DRAFT) -> None:
        if _eq_valid__cls_def is not None:
            self.EQ_VALID__CLS_DEF = _eq_valid__cls_def

        self.init__eq_kwargs(**eq_kwargs)

        self.init__other(other_draft)
        self.init__other_push__chain()

    def init__eq_kwargs(self, **eq_kwargs: TYPING__EQ_VALID__DRAFT) -> None:
        eq_kwargs = eq_kwargs or self.EQ_KWARGS
        self.EQ_KWARGS = DictIcKeys(eq_kwargs)

        if self.EQ_VALID__CLS_DEF is None:
            return

        for key, value in self.EQ_KWARGS.items():
             if not isinstance(value, Base_EqValid):
                self.EQ_KWARGS[key] = self.EQ_VALID__CLS_DEF(value)

    def init__other(self, other_draft: Any | Callable = NoValue) -> None:
        # other_draft ---------------------
        if other_draft is not NoValue:
            self.OTHER_DRAFT = other_draft

        if self.OTHER_FINAL__RESOLVE:
            try:
                self.OTHER_FINAL = Lambda(self.OTHER_DRAFT).resolve__raise()
                self.OTHER_RAISED = False
            except Exception as exc:
                self.OTHER_RAISED = True
                self.OTHER_FINAL = exc
        else:
            self.OTHER_FINAL = self.OTHER_DRAFT

    def init__other_push__chain(self):
        for eq_valid in self.EQ_KWARGS.values():
            if isinstance(eq_valid, Base_EqValid):
                eq_valid.OTHER_FINAL__RESOLVE = False
                eq_valid.OTHER_RAISED = self.OTHER_RAISED

    def _eq_expects__get_final(self, **eq_axpects: bool | None | Any) -> dict[str, bool | None]:
        """
        GOAL
        ----
        two goals:
        1/ if passed any set - get it as pre-final, if not - get default!
        2/ apply lowercase for keys of preFinal
        """
        result = {key.lower(): value for key, value in (eq_axpects or self.EQ_EXPECTS).items()}

        if not result:
            result = {key.lower(): True for key, value in self.EQ_KWARGS.items()}
        return result

    # =================================================================================================================
    def _check_if__(
            self,
            _raise_instead_true: bool = None,
            _iresult_cumulate: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE,
            **eq_axpects: bool | None | Any,
    ) -> bool | NoReturn:
        results: list[bool] = []

        eq_axpects = self._eq_expects__get_final(**eq_axpects)
        for name, expect in eq_axpects.items():

            if name not in self.EQ_KWARGS:
                msg = f"{name=} not in {self.EQ_KWARGS=}"
                raise Exc__WrongUsage(msg)

            if expect is not None:
                result_i = Lambda(
                    lambda: self.EQ_KWARGS[name] == self.OTHER_FINAL,
                ).check_expected__bool(expect)
                results.append(result_i)

        # FINAL result -----------------------
        msg = f"{eq_axpects=}/{results=}"
        print(msg)

        result = Base_ArgsBoolIf(*results, _iresult_cumulate=_iresult_cumulate, _raise_instead_true=_raise_instead_true).resolve()
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def bool_if__all_true(self, **eq_axpects: bool | None | Any) -> bool | NoReturn:
        return self._check_if__(_raise_instead_true=False, _iresult_cumulate=EnumAdj_BoolCumulate.ALL_TRUE, **eq_axpects)

    def bool_if__any_true(self, **eq_axpects: bool | None | Any) -> bool | NoReturn:
        return self._check_if__(_raise_instead_true=False, _iresult_cumulate=EnumAdj_BoolCumulate.ANY_TRUE, **eq_axpects)

    def bool_if__all_false(self, **eq_axpects: bool | None | Any) -> bool | NoReturn:
        return self._check_if__(_raise_instead_true=False, _iresult_cumulate=EnumAdj_BoolCumulate.ALL_FALSE, **eq_axpects)

    def bool_if__any_false(self, **eq_axpects: bool | None | Any) -> bool | NoReturn:
        return self._check_if__(_raise_instead_true=False, _iresult_cumulate=EnumAdj_BoolCumulate.ANY_FALSE, **eq_axpects)

    # -----------------------------------------------------------------------------------------------------------------
    def raise_if__all_false(self, **eq_axpects: bool | None) -> None | NoReturn:
        """
        GOAL
        ----
        useful to check that single POSITIVE(expecting True) variant from any variants (mutually exclusive) expecting TRUE is not correct

        like
            (linux=True, windows=True)
        """
        return self._check_if__(_raise_instead_true=True, _iresult_cumulate=EnumAdj_BoolCumulate.ALL_FALSE, **eq_axpects)

    def raise_if__any_false(self, **eq_axpects: bool | None | Any) -> None | NoReturn:
        """
        GOAL
        ----
        seems like common usage for exact eq-results for special state
            (val1=True, val2=False, val3=True)
        """
        return self._check_if__(_raise_instead_true=True, _iresult_cumulate=EnumAdj_BoolCumulate.ANY_FALSE, **eq_axpects)

    def raise_if__all_true(self, **eq_axpects: bool | None | Any) -> None | NoReturn:
        return self._check_if__(_raise_instead_true=True, _iresult_cumulate=EnumAdj_BoolCumulate.ALL_TRUE, **eq_axpects)

    def raise_if__any_true(self, **eq_axpects: bool | None | Any) -> None | NoReturn:
        return self._check_if__(_raise_instead_true=True, _iresult_cumulate=EnumAdj_BoolCumulate.ANY_TRUE, **eq_axpects)

    # -----------------------------------------------------------------------------------------------------------------
    def __getattr__(self, item: str) -> bool | NoReturn:
        """
        GOAL
        ----
        direct access to final validation (OTHER_DRAFT) by exact item (EqValid) in EQ_KWARGS

        NOTE
        ----
        useful only for existed OTHER_DRAFT
        """
        result = self.bool_if__all_true(**{item: True})
        return result


# =====================================================================================================================
class Base_KwargsEqExpect_StrIc(Base_KwargsEqExpect):
    EQ_VALID__CLS_DEF: type[Base_EqValid] | None = EqValid_EQ_StrIc


# =====================================================================================================================
class KwargsEqExpect_OS(Base_KwargsEqExpect_StrIc):
    OTHER_DRAFT: Any = platform.system()  # NOTE: dont forget lambda *args: !!!! if use callable!

    EQ_KWARGS = dict(
        LINUX="LINUX",
        WINDOWS="WINDOWS",
    )

    # -----------
    WINDOWS: bool
    LINUX: bool


# ---------------------------------------------------------------------------------------------------------------------
class KwargsEqExpect_MachineArch(Base_KwargsEqExpect_StrIc):
    OTHER_DRAFT: Any = platform.machine()

    EQ_KWARGS = dict(
        AMD64="AMD64",        # standard PC
        x86_64="x86_64",      # wsl standard
        AARCH64="AARCH64",    # raspberry=ARM!
    )

    # -----------
    AMD64: bool
    x86_64: bool
    AARCH64: bool


# =====================================================================================================================
if __name__ == "__main__":
    # examples when you need WINDOWS!
    assert KwargsEqExpect_OS().bool_if__any_true(windows=True)
    assert not KwargsEqExpect_OS().raise_if__any_true(linux=True)
    assert KwargsEqExpect_OS().WINDOWS is True
    assert KwargsEqExpect_OS().windows is True


# =====================================================================================================================



import yfinance as yf


def download_stock_data(ticker="AAPL", period='1mo'):
    """
    NOTE
    ----
    cant get SBER UKOIL ...
    """
    stock = yf.Ticker(ticker)
    df = stock.history(period=period)
    return df


if __name__ == "__main__":
    data = download_stock_data()
    print(data)



import copy

from base_aux.base_types.m0_static_typing import TYPING
from base_aux.base_nest_dunders.m1_init1_source import *
from base_aux.base_enums.m2_enum1_adj import *
from base_aux.base_types.m0_static_types import *


# =====================================================================================================================
# @final
class Base_DictAux(NestInit_Source):
    """
    GOAL
    ----
    collect all methods in one class
    """
    SOURCE: TYPING.DICT_ANY_ANY = dict
    SOURCE_INLINE_DEEPCOPY: EnumAdj_SourceOrigOrCopy = EnumAdj_SourceOrigOrCopy.COPY

    # -----------------------------------------------------------------------------------------------------------------
    def _init_post(self) -> None | NoReturn:
        if self.SOURCE_INLINE_DEEPCOPY == EnumAdj_SourceOrigOrCopy.ORIGINAL:
            pass
        if self.SOURCE_INLINE_DEEPCOPY == EnumAdj_SourceOrigOrCopy.COPY:
            self.SOURCE = copy.deepcopy(self.SOURCE)

    # -----------------------------------------------------------------------------------------------------------------
    def values_clear(self) -> TYPING.DICT_ANY_NONE:
        for key in self.SOURCE:
            self.SOURCE[key] = None
        return self.SOURCE

    def values_change__by_func(self, func: Callable[[Any], Any], walk: bool = None) -> TYPING.DICT_ANY_ANY:
        """
        NOTE
        ----
        1/ func work with final/elementary value (not collection) - userClass or Elementary Inst
        2/ func make sure without NoReturn!!!
        3/ FILTERS - use inside func!!! and change values inline!

        GOAL
        ----
        change values by func

        SPECIALLY CREATED FOR
        ---------------------
        json prepare_serialisation - in part of make all values Elementary
        """
        # -----------------------
        for key, value in self.SOURCE.items():
            if isinstance(value, TYPES.ELEMENTARY_COLLECTION):
                if walk:
                    if isinstance(value, dict):
                        value = DictAuxInline(value).values_change__by_func(func, walk=walk)

                    elif isinstance(value, (list, tuple, set)):   # , tuple, set - too complicated!
                        result = []
                        for item in value:
                            if isinstance(item, dict):
                                item = DictAuxInline(item).values_change__by_func(func, walk=walk)
                                # elif isinstance(item, (list, tuple, set)):
                                #     pass
                                #     raise NotImplementedError(f"cant use containers in containers {item}")
                                #     TODO: FINISH - INCOMPLETED!!! simply patched
                            else:
                                item = func(item)    # NOTE: make sure no NoReturn!!!
                            result.append(item)
                        value = result
                else:
                    continue
            else:
                value = func(value)     # NOTE: make sure no NoReturn!!!

            self.SOURCE[key] = value

        return self.SOURCE

    # -----------------------------------------------------------------------------------------------------------------
    def keys_del(self, *keys: Any) -> TYPING.DICT_ANY_ANY:
        for key in keys:
            try:
                self.SOURCE.pop(key)
            except:
                pass

        return self.SOURCE

    def keys_change__by_func(self, func: Callable[[Any], Any], walk: bool = None) -> TYPING.DICT_ANY_ANY:
        """
        GOAL
        ----
        useful to rename keys by func like str.LOWER/upper
        raise on func - delete key from origin! applied like filter ---NO! keep original value!

        SPECIALLY CREATED FOR
        ---------------------
        json prepare_serialisation - in part of make all keys STR!
        """
        # -----------------------
        for key, value in dict(self.SOURCE).items():
            # value -------
            if walk:
                if isinstance(value, dict):
                    value = DictAuxInline(value).keys_change__by_func(func, walk=walk)

                elif isinstance(value, (list, tuple, set)):
                    for item in value:
                        if isinstance(item, dict):
                            item = DictAuxInline(item).keys_change__by_func(func, walk=walk)

            # name -------
            try:
                key_new = func(key)
                self.keys_del(key)
                self.SOURCE[key_new] = value
            except:
                pass

        return self.SOURCE

    # -----------------------------------------------------------------------------------------------------------------
    def key_collapse(self, key: Any) -> TYPING.DICT_ANY_ANY:
        """
        GOAL
        ----
        specially created for 2level-dicts (when values could be a dict)
        so it would replace values (if they are dicts and have special_key)

        CONSTRAINTS
        -----------
        it means that you have similar dicts with same exact keys
            {
                0: 0,
                1: {1:1, 2:2, 3:3},
                2: {1:11, 2:22, 3:33},
                3: {1:111, 2:222, 3:333},
                4: 4,
            }
        and want to get special slice like result

        SPECIALLY CREATED FOR
        ---------------------
        testplans get results for special dut from all results


        main idia to use values like dicts as variety and we can select now exact composition! remain other values without variants

        EXAMPLES
        --------
        dicts like
            {
                1: {1:1, 2:2, 3:3},
                2: {1:1, 2:None},
                3: {1:1},
                4: 4,
            }
        for key=2 return
            {
                1: 2,
                2: None,
                3: None,
                4: 4,
            }

        """
        for root_key, root_value in self.SOURCE.items():
            if isinstance(root_value, dict) and key in root_value:
                self.SOURCE[root_key] = root_value.get(key)

        return self.SOURCE

    # -----------------------------------------------------------------------------------------------------------------
    def prepare_serialisation(self) -> TYPING.DICT_STR_ELEM:
        """
        NOTE
        ----
        work not in source! return copy with ready to direct serialisation keys/values

        GOAL
        ----
        make ready for serialisation
        1/ fix keys - str
        2/ fix values - elementary
        """
        result = {}
        # TODO: FINISH
        result = {}
        for key, value in self.SOURCE.items():
            if isinstance(value, dict):
                value = DictAuxCopy(value).prepare_serialisation()

            if isinstance(value, (list)):
                value = DictAuxCopy(value).prepare_serialisation()

            result[key] = value

        return result

    # -----------------------------------------------------------------------------------------------------------------
    def __str__(self) -> str:
        return self.pretty_str()

    def pretty_str(self, name: str = None, _level: int = 1) -> str:
        """
        GOAL
        ----
        get and print string pretty!
        """
        if name is None:
            name: str = "PRETTY_DICT"
        else:
            name = str(name)

        load = ""

        for key, value in self.SOURCE.items():
            load += f"\n  {key}={value}"

        if not load:
            result = f"{name}={{}}"
        else:
            result = f"{name}={{{load}\n}}"

        print(result)
        return result


# =====================================================================================================================
@final
class DictAuxInline(Base_DictAux):
    SOURCE_INLINE_DEEPCOPY = EnumAdj_SourceOrigOrCopy.ORIGINAL


@final
class DictAuxCopy(Base_DictAux):
    SOURCE_INLINE_DEEPCOPY = EnumAdj_SourceOrigOrCopy.COPY


# =====================================================================================================================
if __name__ == "__main__":
    DictAuxInline({1:1, 2:2}).pretty_str()
    DictAuxInline({}).pretty_str()


# =====================================================================================================================



from typing import *

from base_aux.aux_iter.m1_iter_aux import IterAux
from base_aux.base_values.m2_value_special import *
from base_aux.base_lambdas.m1_lambda import *


# =====================================================================================================================
class DictIcKeys(dict):
    """
    GOAL
    ----
    just a Caseinsense dict
    """
    # __getattr__ = dict.get
    # __setattr__ = dict.__setitem__
    # __delattr__ = dict.__delitem__
    # __iter__ = dict.__iter__
    # __copy__ = dict.copy

    # __repr__ = dict.__repr__  # и так работает!
    # __str__ = dict.__str__    # и так работает!
    # __len__ = dict.__len__    # и так работает!

    # -----------------------------------------------------------------------------------------------------------------
    # GENERIC CLASSES LIKE DICT MUST APPLIED LAST IN MRO!!!
    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)

    # def __init__(self, *args, **kwargs):
    #     _LOCK_KEYS = self.__LOCK_KEYS
    #     super().__init__(*args, **kwargs)
    #     self.__LOCK_KEYS = _LOCK_KEYS

    # -----------------------------------------------------------------------------------------------------------------
    def key__get_original(self, key: str | Any) -> str | NoValue | Any:
        keys_list = list(self)
        result = IterAux(keys_list).item__get_original(key)
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def get(self, item: str | Any) -> Any | None:    # | NoReturn:
        """
        always get value or None!
        if you need check real contain key - check contain)))) [assert key in self] or [getitem_original]
        """
        key_original = self.key__get_original(item)
        if key_original is NoValue:
            return None
            # key_original = item
            # msg = f"{item=}"
            # raise KeyError(msg)

        return super().get(key_original)

    # def set(self, item: Any, value: Any) -> None:

    def pop(self, item: str | Any) -> Any:
        item_original = self.key__get_original(item)
        if item_original is NoValue:
            item_original = item

        return super().pop(item_original)

    def update(self, *args, **kwargs) -> None | NoReturn:
        if args:
            params = args[0]
        else:
            params = kwargs
        for item, value in params.items():
            key_original = self.key__get_original(item)
            if key_original is NoValue:
                key_original = item

            super().update({key_original: value})

    def __contains__(self, item: Any) -> bool:
        return self.key__get_original(item) is not NoValue

    # -----------------------------------------------------------------------------------------------------------------
    # ITEM is universal!
    def __getitem__(self, item: Any) -> Any | NoReturn:
        if item not in self:        # decide use original behaviour or Raise!?
            msg = f"{item=}"
            raise KeyError(msg)
        return self.get(item)

    def __setitem__(self, item: Any, value: Any) -> None:
        self.update({item: value})

    def __delitem__(self, item: Any) -> None:
        self.pop(item)

    # -----------------------------------------------------------------------------------------------------------------
    def __eq__(self, other: dict | Any) -> bool:
        if not isinstance(other, dict):
            return False

        if len(self) != len(other):
            return False

        for key, value in other.items():
            if key not in self or self[key] != value:
                return False

        return True


# =====================================================================================================================
# @final
class DictIc_LockedKeys(DictIcKeys):
    """
    GOAL
    ----
    create INLINE/OnTheFly (ByOneLine) keys set on init object!
    so you dont need create one more/extra object!
    like in
        class MyLockedKeys:
            ATTR1: str = "ATTR1"
            ATTR2: int = 2
            def update(**kwargs):
                pass

        class Final:
            PARAMS = MyLockedKeys()
            def __init__(**kwargs):
                self.PARAMS.update(**kwargs)

    use instead
        class Final:
            PARAMS = DictIc_LockedKeys(ATTR1="ATTR1, ATTR2=2)
            def __init__(**kwargs):
                self.PARAMS.update(**kwargs)

    make dict similar to named tuple
    when you setup keys only on init!

    SPECIALLY CREATED FOR
    ---------------------
    indics and other PARAMS useful objects
    """
    def __setitem__(self, item: Any, value: Any) -> None:
        if item not in self:
            msg = f"[{self.__class__.__name__}]{item=}/{value=}"
            raise Exc__WrongUsage(msg)

        self.update({item: value})

    def update(self, *args, **kwargs) -> None | NoReturn:
        if args:
            params = args[0]
        else:
            params = kwargs
        for item, value in params.items():
            key_original = self.key__get_original(item)
            if key_original is NoValue:
                msg = f"[{self.__class__.__name__}]{item=}/{value=}"
                raise Exc__WrongUsage(msg)

            super().update({key_original: value})


# =====================================================================================================================
# class _DictIcKeys(dict):
#     """
#     DEEP_SEEk +some ref - is not
#     """
#     def __init__(self, arg, **kwargs):
#         super().__init__(arg, **kwargs)
#         self._convert_keys()
#
#     def __setitem__(self, key: Any, value):
#         super().__setitem__(key.lower() if isinstance(key, str) else key, value)
#
#     def __getitem__(self, key: Any):
#         return super().__getitem__(key.lower() if isinstance(key, str) else key)
#
#     def __contains__(self, key: Any):
#         return super().__contains__(key.lower() if isinstance(key, str) else key)
#
#     def get(self, key: Any, default=None):
#         return super().get(key.lower() if isinstance(key, str) else key, default)
#
#     def pop(self, key: Any, default=None):
#         return super().pop(key.lower() if isinstance(key, str) else key, default)
#
#     def update(self, other=None, **kwargs):
#         if other is not None:
#             if hasattr(other, 'items'):
#                 for key, value in other.items():
#                     self[key] = value
#             else:
#                 for key, value in other:
#                     self[key] = value
#         for key, value in kwargs.items():
#             self[key] = value
#
#     def _convert_keys(self):
#         for key in list(self.keys()):
#             if isinstance(key, str):
#                 value = super().pop(key)
#                 self[key] = value


# =====================================================================================================================



import pathlib

from base_aux.base_nest_dunders.m3_calls import *
from base_aux.aux_attr.m2_annot3_gacls_keys_as_values import *
from base_aux.path0_static.m1_file_extensions import FilesStdExtensions


# =====================================================================================================================
@final
class Resolve_DirPath(NestInit_Source, NestCall_Resolve):
    """
    GOAL
    ----
    resolve dirpath by draft
    if file and fileExist - return parent!

    SPECIALLY CREATED FOR
    ---------------------
    Resolve_FilePath init dirpath
    """
    SOURCE: TYPING.PATH_DRAFT

    def resolve(self) -> TYPING.PATH_FINAL:
        # ---------------------
        if self.SOURCE is None:
            return pathlib.Path().cwd()

        # ---------------------
        if self.SOURCE is not None:
            self.SOURCE = pathlib.Path(self.SOURCE)

        self.SOURCE: pathlib.Path

        # try detect files by existed ---------------------
        if self.SOURCE.exists() and self.SOURCE.is_file():
            return self.SOURCE.parent

        # try detect files by extensions ---------------------
        splited = self.SOURCE.name.rsplit(".")
        if len(splited) == 2:
            name, extlast = splited
            if extlast in FilesStdExtensions:
                return self.SOURCE.parent

        # FINAL ---------------------
        return self.SOURCE


# =====================================================================================================================



# =====================================================================================================================
# STARICHENKO UNIVERSAL IMPORT
import sys
sys.path.append("..")  # Adds higher directory to python modules path.

import time
import pathlib
import copy
import threading


# SINGLETON ===========================================================================================================
# 1=NEW======= USE!
class Meta_Singleton(type):   # starichenko
    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, '__INSTANCE'):
            # FIRST VARIANT -------------------- cant create independent singletones!
            # if not hasattr(cls, 'INSTANCE'):
            #     cls.INSTANCE = super().__call__(*args, **kwargs)
            # return cls.INSTANCE

            # SECOND VARIANT --------------------
            # you need create attribute otherwise sometimes (DWDM) infinity resursive!!!
            # cls.__INSTANCE = None   # this is dont work!!!
            setattr(cls, '__INSTANCE', None)    # NEED THIS!!!

            cls.__INSTANCE = super().__call__(*args, **kwargs)

        return cls.__INSTANCE


# 2=OLD======= DONT USE!!!!
class _Singleton:   # starichenko
    """
    Create singleton class by inheriting this one!
    DONT CRIATE OBJECT DIRECT from this class!!!

    IMPORTANT STEPS:
    1=inherit your class from this one!
    2=in inheritance order place this class to last place!!!
    3=dont forget decorating __INIT__ by decorator_singleton_for_class_method_init!!!
        otherwise init will be called and all instances (links) will be owerinited to init state!!!

    see tests for example!
    """
    # TODO in future you can create single CLASS_DECORATOR! (if possible)
    # _INSTANCE = None              # moved to child class!!!
    # _INSTANCE_INITED_FLAG = None  # moved to child class!!!

    def __new__(cls, *args, **kwargs):
        # print(f"{args=}{kwargs=}")
        child_class = cls.__mro__[0]

        # you need to use inheritance from this class! Not direct!
        if __class__ == cls:
            msg = f"""
                    this class must not run directly!!!
                    use at least simple sample (its necessary and enough!)
                    class MyNewClassSingleton({cls.__name__}):
                        pass
            """
            raise Exception(msg)

        if not hasattr(child_class, "_INSTANCE") or child_class._INSTANCE is None:
            child_class._INSTANCE = super().__new__(cls)
        return child_class._INSTANCE


def _decorator_singleton_for_class_method_init(_func_link):  # starichenko
    def _wrapper_for_class_method(self, *args, **kwargs):
        # print(f"decorating[{self.__class__.__name__}.{_func_link.__name__}()]")
        if not hasattr(self.__class__, "_INSTANCE_INITED_FLAG") or not self.__class__._INSTANCE_INITED_FLAG:
            self.__class__._INSTANCE_INITED_FLAG = True
            return _func_link(self, *args, **kwargs)
        else:
            for class_i in self.__class__.__mro__:
                # in case of reinicialization _DictDotAttrAccess2_Singleton with new params!
                if class_i.__name__ == "_DictDotAttrAccess2_Singleton":
                    return class_i.__init__(self, *args, **kwargs)
            return None
    return _wrapper_for_class_method


# =====================================================================================================================
class _DictDotAttrAccess2_Singleton(_Singleton):    # starichenko
    """
    ***i tried create nonSingleton! but get fail!

    IMPORTANT:
        HERE INSIDE:
            DONT TOUCH/CHANGE ANY LINES WITHOUT IMMEDIATE CHECK BY TESTING!!! very easy to break it!!!

    ------------------------------------
    singleton alternative complicated variant for standard dict!

    differences from simple DICT
    0. all standard methods are usefull! (if not dont forget to add with tests!!)
    1. Singleton
    1. exists some another custom usefull methods!
    2. dotted attribute access! (only for first level)
    3. work with json file (can autosave dict)

    4. safe for input source dict!!! no inline changes!

    ------------------------------------
    USAGE
    create SINGLETONS:
        1. inherit this class (even if you need only one DictSingleton)!
            class Cls1(_DictDotAttrAccess2_Singleton):
                pass
            class Cls2(_DictDotAttrAccess2_Singleton):
                pass

        2. DONT REINHERIT ELSE ONE new your classes!!!
            class Cls3(Cls2):   # cause incorrect work! use only as shown above!!!
                pass

    add new ATTRIBUTES
        3. ALL ATTRIBUTES MUST BE FIRST INITED AS CLASS ATTR!!! only after this step you can use it by SELF reference!!!
            class Cls1(_DictDotAttrAccess2_Singleton):
                hello = None
                def hello_set(self, val):
                    self.hello = val

    add INIT
        4. if need init - DONT FORGET START super().__init__(*args, **kwargs) DEFINITLY AT FIRTS LINE!!!
        BUT i think it is not recommended to use INIT!!! try not use!
        but maybe you need some some reload from file?

            class Cls1(_DictDotAttrAccess2_Singleton):
                hello = 1
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.hello = 111

        5. if you need singleton init use @_decorator_singleton_for_class_method_init
        if no (need create yours attr on init state - dont use it BUT IT IS) dont use! but dict data will stay singleton!

    add new METHODS
        6. add it as usual!

    -------------------------------------
    РАБОТА С АТТРИБУТАМИ обьекта/класса
    0. теперь можно обращаться как к классовым атрибутам так и к обьектовым!
    - все равно будет изменяться и выдаваться именно классовый!!!
    1. если класс имеет атрибут - то любое обращение идет по нему!
    2. если класс не имеет атрибут, то любое обращение к его имени подразумевает обращение к словарю _DICT!
    - тоесть если значения там нет - то читается None, а при записи создается там!

    fully created by starichenko
    """

    # TODO: create minimal or even stable keys list for DICT ??????
    # RECOMMENDED ALL CLASS ATTR NAMES KEEP UPPERCASE!
    # AND DONT HIDE TO [__*] - will not work on inheritance!
    # _DICT = {}                # moved to child class!!!
    # _FILEPATH = None          # moved to child class!!!
    # _AUTO_DUMP_FLAG = True    # moved to child class!!!

    # get = dict.get    # dont work! dont know why!

    # DONT DECORATE BY SINGLETON!!! useful to keep it opened!
    def __init__(self, data=None, filepath=None, auto_dump_flag=None):     # CAREFULL! apply _DICT only if you want to ovewrite all instances!
        """

        ANY PARAMS REPLACE IN SINGLETON ONLY IF PASSED IN INIT!
        if you want to clear some of them - use special methods! or pass any value except NONE!

        :param dict:
        :param filepath: if you have it - dict will be automatic dumped on any data changed!
        :param auto_dump_flag:
        """
        super().__init__()
        self.__init_if_not_singleton_child_class_base_attributes()

        if isinstance(data, dict):
            self.replace_dict(data)
        elif data is None:
            pass
        else:
            msg = f"incorrect input [{data=}]"
            print(msg)

        if filepath is not None:
            self.filepath_change(filepath)

        if auto_dump_flag is not None:
            self.auto_dump_flag_change(auto_dump_flag)

    # BASE SINGLETON PREPARE ==========================================================================================
    def __init_if_not_singleton_child_class_base_attributes(self):
        """
        create base attr in child class
        :return:
        """
        if not hasattr(self.__class__, "_DICT"):
            self.__class__._DICT = {}

        if not hasattr(self.__class__, "_FILEPATH"):
            self.__class__._FILEPATH = None

        if not hasattr(self.__class__, "_AUTO_DUMP_FLAG"):
            self.__class__._AUTO_DUMP_FLAG = True

    # ATTRIBUTES WORK ======================
    def get_dict(self, get_safe_copy=True):
        """
        recommended use exact get_safe_copy!!! else you will have direct access to base DICT!!!!
        :param get_safe_copy:
        :return:
        """
        if get_safe_copy:
            return copy.deepcopy(self._DICT)
        else:
            return self._DICT

    def __getattr__(self, item):
        if item in dir(self):
            return getattr(self.__class__, item)
        else:
            return self.__class__._DICT.get(item)

    def __setattr__(self, key, value):
        if key in dir(self):
            # print()
            # print(self.__class__)
            setattr(self.__class__, key, value)
        else:
            self.update({key: value})

    def __delattr__(self, item):
        if item in dir(self):
            delattr(self.__class__, item)
        else:
            self.delete(item)

    # MAGIC STAFF ======================
    def __bool__(self):
        return bool(self._DICT)

    def __str__(self):
        return str(self._DICT)

    def __repr__(self):
        return repr(self._DICT)

    def __len__(self):
        return len(self._DICT)

    def __getitem__(self, key):
        return self._DICT.get(key)

    def __setitem__(self, key, value):
        self.update({key: value})

    def __delitem__(self, key):
        self.delete(key)

    def __iter__(self):
        return iter(self._DICT)

    def __eq__(self, other):
        return self._DICT == other

    def __ne__(self, other):
        return self._DICT != other

    # have NOT IN MAGIC staff ======================
    def get(self, key, _default=None):      # DONT RENAME!!! it uses in dict!!!
        return self._DICT.get(key, _default)

    def items(self):
        return self._DICT.items()

    def keys(self):
        return self._DICT.keys()

    def values(self):
        return self._DICT.values()

    def clear(self):
        self.delete()

    def update(self, new_dict={}):          # LINK all CHANGES to here!!!!
        self._DICT.update(new_dict)
        self.file_dump_if_auto()

    # CUSTOM (have not in DICT) ======================
    def set(self, key, value=None):
        self.update({key: value})

    def delete(self, key=None):             # LINK all DELETES to here!!!!
        if key is None:
            self._DICT.clear()
        else:
            try:
                del self._DICT[key]
            except:
                pass
        self.file_dump_if_auto()

    def check_empty(self):
        return self._DICT == {}

    def replace_dict(self, new_dict={}):    # LINK all REPLACES to here!!!!
        """
        fully replace original dict!
        you can use simply reinit new instance with this dict!

        IMPORTANT: DONT CHANGE TO SIMPLY
                self._DICT = new_dict
        IT WILL BREAK ALL WORKING!!!

        :param new_dict:
        :return:
        """
        self.clear()
        self._DICT.update(new_dict)
        self.file_dump_if_auto()

    # filepath ------------------------------------------------------------------------------------------------------
    def auto_dump_flag_change(self, auto_dump_flag=None):
        self._AUTO_DUMP_FLAG = auto_dump_flag

    def filepath_change(self, filepath=None):
        if filepath:
            self._FILEPATH = pathlib.Path(filepath)
        else:
            self._FILEPATH = None

    def file_dump_if_auto(self):
        # print(111)
        if self._AUTO_DUMP_FLAG:
            self.file_bump()

    def file_bump(self, filepath=None):
        if filepath:
            filepath = pathlib.Path(filepath)
        else:
            filepath = self._FILEPATH

        if filepath:
            return ProcessorJsonDict().json_dump(filepath=filepath, json_dict=self.get_dict())

    def file_load(self, filepath=None):
        if filepath:
            filepath = pathlib.Path(filepath)
        else:
            filepath = self._FILEPATH

        if filepath:
            data = ProcessorJsonDict().json_read(filepath)
            self.replace_dict(data)
            return data


# TESTS ===============================================================================================================
def test__Singleton__decorator_singleton_for_class_method_init():
    test_obj_link1 = _Singleton
    test_obj_link2 = _decorator_singleton_for_class_method_init

    class Cls1(test_obj_link1):
        cls_atr = 1
        @test_obj_link2
        def __init__(self, hello):
            self.atr = hello

    class Cls2(test_obj_link1):
        cls_atr = 2
        @test_obj_link2
        def __init__(self, hello):
            self.atr = hello

    def start_in_thread():
        obj_thread = Cls1(1)
        for i in range(10):
            print(f"{obj_thread.atr=}")
            time.sleep(0.5)

    thread = threading.Thread(target=start_in_thread, daemon=1)
    thread.start()
    time.sleep(1)

    print()
    obj1 = Cls1(1)
    print(F"{obj1.atr=}")
    assert obj1.atr == 1

    obj1.atr = 2
    print(F"{obj1.atr=}")
    assert obj1.atr == 2

    # 2=check not initing new instances!
    print()
    time.sleep(1)
    obj2 = Cls1(1)
    print(F"{obj1.atr=}")
    print(F"{obj2.atr=}")
    assert obj1.atr == 2
    assert obj2.atr == 2
    print()

    time.sleep(2)

    # check TWO CLASSES inherited from one _Singleton!!!
    # 1=only self attr!
    obj3 = Cls2(3)
    print(F"{obj1.atr=}")
    print(F"{obj3.atr=}")
    assert obj1.atr == 2
    assert obj3.atr == 3

    # 2=class attr - ОТВЯЗАННЫЙ!!! тоесть неизменяемый внутри!!!
    print()
    print()
    print(F"{Cls1().cls_atr=}")
    assert Cls1().cls_atr == 1
    Cls1().cls_atr = 111
    print(F"{Cls1().cls_atr=}")
    assert Cls1().cls_atr == 111

    print()
    print(F"{Cls2().cls_atr=}")
    assert Cls2().cls_atr == 2
    Cls2().cls_atr = 222
    print(F"{Cls2().cls_atr=}")
    assert Cls2().cls_atr == 222

    print(F"{Cls1().cls_atr=}")
    assert Cls1().cls_atr == 111

    Cls1().atr = 1
    assert Cls1().atr == 1
    assert Cls2().atr == 3

    # 3=class attr - ИЗМЕНЯЕМЫЙ внутри!!!


def test__DictDotAttrAccess2_Singleton():  # starichenko
    # test_obj_link = UFU._DictDotAttrAccess2_Singleton
    class test_obj_link(_DictDotAttrAccess2_Singleton):
        pass

    # SINGLETON -----------------------------------------------------------------------
    test_obj_link({1: 1})
    assert len(test_obj_link()) == 1  # SINGLETON
    assert len(test_obj_link({})) == 0  # SINGLETON

    obj1 = test_obj_link({1: 1})
    obj2 = test_obj_link({2: 2})
    assert obj1 == {2: 2}
    obj2[2] = 222
    assert obj1 == {2: 222}

    # INHERIT --------------------------------------------------------------------------
    # class Cls1(copy.copy(UFU._DictDotAttrAccess2_Singleton)):      # не помогло!!!
    class Cls1(_DictDotAttrAccess2_Singleton):
        hello = None
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.hello = 1
    class Cls2(_DictDotAttrAccess2_Singleton):
        hello = None
        @_decorator_singleton_for_class_method_init
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.hello = 2

    # several indepandent classes -----------------------------------------------
    assert test_obj_link({1:1}) == {1:1}
    assert Cls1() == {}                     # indepandent cls 1
    Cls1().attr = 111
    assert Cls1() == {"attr": 111}
    assert Cls2() != {"attr": 111}          # indepandent cls 2

    assert Cls1({2: 2}) == Cls2({2: 2})
    assert Cls1({}) != Cls2({3:3})

    # INITs
    assert Cls1({1:1}).hello == 1
    Cls1().hello = 111
    assert Cls1().hello == 1  # no singleton init

    assert Cls2({2:2}).hello == 2
    Cls2().hello = 222
    assert Cls2().hello == 222     # singleton init

    # ATTRIBUTES -----------------------------------------------------------------------
    # attr DICT
    assert test_obj_link({"atr": 1}) == {"atr": 1}            # .atr
    assert test_obj_link().atr == 1
    test_obj_link().atr = 111
    assert test_obj_link().atr == 111
    del test_obj_link().atr                # del.attr
    assert test_obj_link().atr == None
    del test_obj_link().atr                # del.attr not exists

    # attr OBJECT
    assert Cls1({1:1}).hello == 1
    assert Cls2({}).hello == 222
    assert Cls1({}) == {}
    assert Cls2({}) == {}

    # attr CLASS
    assert test_obj_link({})._DICT == {}
    test_obj_link().atr = 111
    assert test_obj_link()._DICT == {"atr": 111}

    assert test_obj_link._AUTO_DUMP_FLAG == True
    assert test_obj_link()._AUTO_DUMP_FLAG == True
    test_obj_link()._AUTO_DUMP_FLAG = False
    assert test_obj_link._AUTO_DUMP_FLAG == False
    assert test_obj_link()._AUTO_DUMP_FLAG == False
    assert test_obj_link()._DICT == {"atr": 111}

    assert test_obj_link({1:1}, auto_dump_flag=False)._AUTO_DUMP_FLAG == False
    assert test_obj_link(auto_dump_flag=True)._AUTO_DUMP_FLAG == True
    assert test_obj_link()._AUTO_DUMP_FLAG == True
    assert test_obj_link()._DICT == {1: 1}

    # REPRESENTS -----------------------------------------------------------------------
    assert str(test_obj_link({})) == "{}"     # STR
    assert str(test_obj_link({1: 1})) == "{1: 1}"

    assert bool(test_obj_link({})) == False           # BOOL
    assert bool(test_obj_link({1:1})) == True

    # ITER ----------------------------------------------------------------------------
    assert len(test_obj_link({1: 1})) == 1                # LEN
    assert list(test_obj_link({1: 1})) == [1]             # LIST

    new_list = []
    for key in test_obj_link():  # ITER
        new_list.append(key)
    assert list(test_obj_link()) == new_list

    assert (1 in test_obj_link({1: 1})) == True   # IN
    assert (111 in test_obj_link({1: 1})) == False   # IN

    new_dict = {}
    for key, value in test_obj_link().items():                      # ITEMS
        new_dict.update({key: value})
    assert list(test_obj_link().keys()) == list({1:1}.keys())       # keys
    assert list(test_obj_link().values()) == list({1:1}.values())   # values

    assert (test_obj_link() == new_dict) == True    # EQ
    assert (test_obj_link() != new_dict) == False   # NE

    # GET/SET/DEL --------------------------------------------------------------------------
    test_obj_link({}).set(11, 11)
    assert test_obj_link() == {11: 11}              # set(key, value)

    assert test_obj_link({1: 1}).get(1) == 1              # get()
    assert test_obj_link().get(1, 999) == 1         # get() default
    assert test_obj_link().get(2, 999) == 999       # get() default

    assert test_obj_link()[1] == 1                  # get[]
    assert test_obj_link()[2] == None               # get[] default

    test_obj_link()[1] = 11
    assert test_obj_link()[1] == 11         # set[]

    del test_obj_link()[1]
    del test_obj_link()[1]                  # del[] not exists
    assert test_obj_link()[1] == None       # del[]

    # SAFE SOURCE DICT ACCESS ---------------------------------------------------------
    temp_dir = {1:1}
    test_obj_link(temp_dir)
    assert test_obj_link()[1] == 1      # SAFE SOURCE DICT
    del test_obj_link()[1]
    assert test_obj_link()[1] == None
    assert temp_dir == {1:1}

    temp_dir = test_obj_link({})._DICT
    assert temp_dir == {}
    assert test_obj_link() == {}
    test_obj_link()[1] = 1
    assert temp_dir == {1:1}
    temp_dir[1] = 111
    assert test_obj_link() == {1:111}        # NOT SAFE ACCESS!

    temp_dir = test_obj_link({}).get_dict(get_safe_copy=False)
    assert temp_dir == {}
    assert test_obj_link() == {}
    test_obj_link()[1] = 1
    assert temp_dir == {1:1}
    temp_dir[1] = 111
    assert test_obj_link() == {1:111}        # NOT SAFE ACCESS!

    temp_dir = test_obj_link({}).get_dict(get_safe_copy=True)
    assert temp_dir == {}
    assert test_obj_link() == {}
    test_obj_link()[1] = 1
    assert temp_dir == {}
    temp_dir[1] = 111
    assert test_obj_link() == {1:1}        # SAFE ACCESS!

    # FULL DICT ------------------------------------------------------------------------
    assert test_obj_link().get_dict() == {1:1}         # get_dict
    # assert dict(test_obj_link()) == {1:1}         # dict()  --------НЕ РАБОТАЕТ!!! и не особо важно!

    assert test_obj_link().check_empty() == False  # check_empty

    test_obj_link({1:1}).clear()
    assert test_obj_link().check_empty() == True    # clear

    test_obj_link().update({1:1, 2:2})
    assert test_obj_link() == {1:1, 2:2}    # UPDATE
    test_obj_link().update({1:111})
    assert test_obj_link() == {1:111, 2:2}

    test_obj_link({1:1}).replace_dict({11: 11})
    assert test_obj_link() == {11:11}         # replace_dict

    # FILE ------------------------------------------------------------------------
    assert test_obj_link({1:1}, auto_dump_flag=True) == {1:1}
    assert test_obj_link()._AUTO_DUMP_FLAG == True
    test_obj_link().auto_dump_flag_change(False)
    assert test_obj_link()._AUTO_DUMP_FLAG == False     # auto_dump_flag_change

    test_obj_link().filepath_change("hello.json")
    assert test_obj_link()._FILEPATH == pathlib.Path("hello.json")     # filepath_change

    test_obj_link({1:1}).file_bump("hello.json")


# =====================================================================================================================



from base_aux.versions.m2_version import *
from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m3_exceptions import *

from base_aux.tables.m1_table_obj import TableObj
from base_aux.tables.m2_table_column import TableColumn


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="data, col, name, _EXPECTED",
    argvalues=[
        # TableObj -------------
        (dict(), 0, "ATC", Exception),

        (dict(ATC=[]), 0, "ATC", Exception),
        (dict(ATC=[]), 1, "ATC", Exception),

        (dict(ATC=[1,]), 0, "ATC", 1),
        (dict(ATC=[1,]), 1, "ATC", Exception),
        (dict(ATC=[1,]), 0, "atc", Exception),

        (dict(ATC=[1,2,]), 1, "ATC", 2),

        (dict(ATC=[1,], PTB=[2,]), 0, "PTB", 2),
        (dict(ATC=[1,], PTB=(2,)), 0, "PTB", 2),
        (dict(ATC=[1,], PTB={2,}), 0, "PTB", Exception),

        (dict(ATC="atc", PTB=[1, 2, 3]), 2, "ATC", "c"),
    ]
)
def test__full(data, col, name, _EXPECTED):
    for victim in [data, TableObj(**data)]:
        Lambda(lambda: TableColumn(table=victim, column=col)[name]).check_expected__assert(_EXPECTED)
        Lambda(lambda: getattr(TableColumn(table=victim, column=col), name)).check_expected__assert(_EXPECTED)


# =====================================================================================================================




from base_aux.aux_argskwargs.m2_argskwargs_aux import *
from base_aux.base_types.m0_static_types import *
from base_aux.base_types.m0_static_typing import TYPING


# =====================================================================================================================
class ClsMiddleGroup:
    """
    # FIXME: need to deprecate? no! YES? use EqCls

    NOTE1: DONT deprecate! Cant compare methods even classmethods - compare only aux_attr!
    ------------------------
    1. comparing direct methods on base_types will not work!!!
        -------------------------
            class Cls:
                def meth(self):
                    pass
                @classmethod
                def cmeth(cls):
                    pass

            class Cls2(Cls):
                pass

            print(Cls.meth == Cls2.meth)        # True
            print(Cls().meth == Cls2().meth)    # false
            print(Cls().cmeth == Cls2().cmeth)  # false
        -------------------------

    2. and even cmp classes - will not work!
        -------------------------
            class Cls:
                def meth(self):
                    pass
                @classmethod
                def cmeth(cls):
                    pass

            class Cls2(Cls):
                pass

            print(Cls.meth is Cls2.meth)
            print(Cls.meth == Cls2.meth)
            print(Cls.cmeth is Cls2.cmeth)
            print(Cls.cmeth == Cls2.cmeth)

            print(Cls.meth)
            print(Cls2.meth)
            print(Cls.cmeth)
            print(Cls2.cmeth)

            #
            True
            True
            False
            False
            <function Cls.meth at 0x000002CC6B525940>
            <function Cls.meth at 0x000002CC6B525940>
            <bound method Cls.cmeth of <class '__main__.Cls'>>
            <bound method Cls.cmeth of <class '__main__.Cls2'>>
        -------------------------

    NOTE2: under your own responsibility
    ------------------------------------
    you need keep separating groups by yourself!

    CREATED SPECIALLY FOR
    ---------------------
    module testplans
    need to handle testcase classes as some groups!
    there was not enough separating process just by startup_cls and startup_inst!!! need startup_group!

    THERE ARE THREE WAYS TO SEPARATE GROUPS
    ---------------------------------------
    1. ONLY NAME
    str attribute as base (always checks first) - MIDDLE_GROUP__NAME
        class Group1(ClsMiddleGroup):
            MIDDLE_GROUP__NAME = "Group1"

        class Group2(ClsMiddleGroup):
            MIDDLE_GROUP__NAME = "Group2"

    2. ONLY ATTR VALUES - MIDDLE_GROUP__CMP_ATTR
        class GroupBase(ClsMiddleGroup):
            MIDDLE_GROUP__CMP_ATTR = "meth"
            MIDDLE_GROUP__CMP_ATTR = ["attr", "attr2"]

        class Group1(GroupBase):
            attr = 1

        class Group2(GroupBase):
            attr = 2


    3. BOTH NAME/ATTRS






    further info may be old and incorrect!
    ======================================
    GOAL
    ----
    if you need to separate cls/inst into several groups by middle nested class.
    main idea is manipulating with grouped methods, not just name in attribute.

    USAGE
    -----
    1. create some middle classes which will define groups.
    2. apply group methods and mention it in MIDDLE_GROUP__CMP_ATTR so base_types could be compared not just by names!
    3. do nesting your final classes by Groups

        # BEFORE --------------------------
        class YourClsBase:
            pass

        class Cls1(YourClsBase):
            GROUP__NAME = "Group1"
            def meth(self):
                pass

        class Cls2(YourClsBase):
            GROUP__NAME = "Group2"
            def meth(self):
                return True

        class Cls3(YourClsBase):
            GROUP__NAME = "Group2"
            def meth(self):
                return True

        # AFTER --------------------------
        class Group1(ClsMiddleGroup):
            MIDDLE_GROUP__NAME = "Group1"
            def meth(self):
                pass

        class Group2(ClsMiddleGroup):
            MIDDLE_GROUP__NAME = "Group2"
            def meth(self):
                return True

        class YourClsBase(ClsMiddleGroup):
            pass

        class Cls1(Group1, YourClsBase):
            pass

        class Cls2(Group2, YourClsBase):
            pass

        class Cls3(Group2, YourClsBase):
            pass
    """
    MIDDLE_GROUP__NAME: None | str = None           # main cmp meth
    MIDDLE_GROUP__CMP_ATTR: TYPING.ARGS_FINAL = None       # additional cmp parameters

    @classmethod
    def middle_group__check_equal__cls(cls, other: Any) -> bool | None:
        """
        middle groups could be cmp only by this method!
        """
        if not TypeAux(other).check__subclassed_or_isinst__from_cls_or_inst(ClsMiddleGroup):
            return

        other = TypeAux(other).get__class()

        # in case of used only empty base MiddleGroup and no any Group configured!
        if cls.MIDDLE_GROUP__NAME == other.MIDDLE_GROUP__NAME == cls.MIDDLE_GROUP__CMP_ATTR == other.MIDDLE_GROUP__CMP_ATTR is None:
            return False

        # CMP not only by just one name!!! need cmp by equity special methods!
        if cls.MIDDLE_GROUP__NAME != other.MIDDLE_GROUP__NAME:
            return False

        attrs = []
        if cls.MIDDLE_GROUP__CMP_ATTR:
            attrs1 = ArgsKwargsAux(cls.MIDDLE_GROUP__CMP_ATTR).resolve_args()
            attrs.extend(attrs1)

        if other.MIDDLE_GROUP__CMP_ATTR:
            attrs2 = ArgsKwargsAux(other.MIDDLE_GROUP__CMP_ATTR).resolve_args()
            attrs.extend(attrs2)

        # cmp ------------------
        for attr in attrs:
            try:
                getattr1 = getattr(cls, attr, None)
                getattr2 = getattr(other, attr, None)
                if getattr1 != getattr2:
                    return False
            except:
                return False

        return True

    def middle_group__check_equal__inst(self, other: Any) -> bool | None:
        """
        NOT REALIZED YET! and maybe not available!
        """
        pass

    # -----------------------------------------------------------------------------------------------------------------
    # EXAMPLE for TCS
    # def startup__group(self):
    #     pass
    # def teardown__group(self):
    #     pass


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_types.m2_info import *
from base_aux.htmls.m1_html_tag import *


# =====================================================================================================================
EXAMPLE_HTML = """
<!doctype html>
<html lang="en-US">
  <head>
    <meta charset="utf-8">
    <title>TITLE</title>
  </head>
  <body>
    <p>TextP1</p>
    <p>TextP2</p>
    <a href="http://example.com/">LinkText<i>Italic</i></a>
    <a cls="cls1" href="http://example.com/link1" id="link1">TextLink1</a>
  </body>
</html>
"""


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, address, _EXPECTED",
    argvalues=[
        (EXAMPLE_HTML, dict(name="p"), "TextP1"),
        (EXAMPLE_HTML, dict(name="p", index=0), "TextP1"),
        (EXAMPLE_HTML, dict(name="p", index=1), "TextP2"),
        (EXAMPLE_HTML, dict(name="p", index=2), None),
        (EXAMPLE_HTML, dict(name="a", attrs=dict(cls="cls1")), "TextLink1"),
        (EXAMPLE_HTML, dict(name="a", attrs=dict(id="link1")), "TextLink1"),
        (EXAMPLE_HTML, dict(name="a123", attrs=dict(id="link1")), None),
    ]
)
def test__tag(source, address, _EXPECTED):
    Lambda(HtmlTagParser(source, **address).resolve).check_expected__assert(_EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, chain_dicts, _EXPECTED",
    argvalues=[
        # one chain ------
        (EXAMPLE_HTML, [dict(name="p"),], "TextP1"),
        (EXAMPLE_HTML, [dict(name="p", index=0),], "TextP1"),
        (EXAMPLE_HTML, [dict(name="p", index=1),], "TextP2"),
        (EXAMPLE_HTML, [dict(name="p", index=2),], None),
        (EXAMPLE_HTML, [dict(name="a", attrs=dict(cls="cls1")),], "TextLink1"),
        (EXAMPLE_HTML, [dict(name="a", attrs=dict(id="link1")),], "TextLink1"),
        (EXAMPLE_HTML, [dict(name="a123", attrs=dict(id="link1")),], None),

        # reveral chains ------
        (EXAMPLE_HTML, [dict(name="body"), dict(name="p", index=1)], "TextP2"),
    ]
)
def test__chain(source, chain_dicts, _EXPECTED):
    Lambda(ChainResolve_HtmlTagParser(*[HtmlTagParser(**chain_dict) for chain_dict in chain_dicts], source=source).resolve).check_expected__assert(_EXPECTED)


# =====================================================================================================================
def test__direct():
    victim = HtmlTagParser(source=EXAMPLE_HTML, name="p")
    print(victim.resolve())

    body = HtmlTagParser(source=EXAMPLE_HTML, name="body").resolve()
    print(f"{body=}")

    assert HtmlTagParser(source=body, name="p").resolve() == "TextP1"

    # Lambda(victim.resolve).expect__check_assert("TextP1")
    # pass


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class PROJECT:
    # AUTHOR -----------------------------------------------
    AUTHOR_NAME: str = "Andrei Starichenko"
    AUTHOR_EMAIL: str = "centroid@mail.ru"
    AUTHOR_HOMEPAGE: str = "https://github.com/centroid457/"

    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "server_templates"
    KEYWORDS: List[str] = [
        "api", "api server", "http server",
        "requests",
        "aiohttp",
        "FastApi",
    ]
    CLASSIFIERS_TOPICS_ADD: List[str] = [
        # "Topic :: Communications",
        # "Topic :: Communications :: Email",
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "templates for servers"
    DESCRIPTION_LONG: str = """designed for keep all servers templates in one place"""
    FEATURES: List[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        ["[SERVERS]",
            "[aiohttp] (try not to use, as old)",
            "[FastApi] (preferred)",
         ],
        "client_requests item+stack",
    ]

    # HISTORY -----------------------------------------------
    VERSION: Tuple[int, int, int] = (0, 3, 3)
    TODO: List[str] = [
        "...",
    ]
    FIXME: List[str] = [
        "...",
    ]
    NEWS: List[str] = [
        "[clientRequestsItem] fix check_success",

        "[__INIT__.py] fix import",
        "apply last pypi template",
    ]


# =====================================================================================================================



import pytest
from typing import *

from base_aux.pyqt.m0_signals import *
from PyQt5.QtCore import pyqtSignal


# =====================================================================================================================
class Test__Signals:
    # -----------------------------------------------------------------------------------------------------------------
    def setup_method(self, method):
        class Signals(Base_Signals):
            signal__send_no = pyqtSignal()

            signal__send_one__wo_def = pyqtSignal(int)
            signal__send_one__wo_def__not_used = pyqtSignal(int)
            signal__send_one__with_def = pyqtSignal(int)

            signal__send_several__wo_def = pyqtSignal(int, int)
            signal__send_several__with_def2 = pyqtSignal(int, int)
            signal__send_several__with_def12 = pyqtSignal(int, int)

            signal__send_any = pyqtSignal(Any)

        class Victim:
            SIGNALS = Signals()

            value: Union[None, int, list[int]] = None

            def __init__(self):
                self.slots_connect()

            def slots_connect(self):
                self.SIGNALS.signal__send_no.connect(self.clear)

                self.SIGNALS.signal__send_one__wo_def.connect(self.value__set_one__wo_def)
                self.SIGNALS.signal__send_one__wo_def__not_used.connect(self.value__set_one__wo_def__not_used)
                self.SIGNALS.signal__send_one__with_def.connect(self.value__set_one__with_def)

                self.SIGNALS.signal__send_several__wo_def.connect(self.value__set_several__wo_def)
                self.SIGNALS.signal__send_several__with_def2.connect(self.value__set_several__with_def2)
                self.SIGNALS.signal__send_several__with_def12.connect(self.value__set_several__with_def12)

            def clear(self):
                self.value = None

            # -----------------------------------------------
            def value__set_one__wo_def(self, value):
                self.value = value

            def value__set_one__wo_def__not_used(self, value):
                self.value = 999

            def value__set_one__with_def(self, value=None):
                self.value = value

            # -----------------------------------------------
            def value__set_several__wo_def(self, value1, value2):
                self.value = [value1, value2]

            def value__set_several__with_def2(self, value1, value2=None):
                self.value = [value1, value2]

            def value__set_several__with_def12(self, value1=None, value2=None):
                self.value = [value1, value2]

        self.victim = Victim()

    # -----------------------------------------------------------------------------------------------------------------
    def test__several_emits(self):
        victim = self.victim

        victim.SIGNALS.signal__send_no.emit()
        victim.SIGNALS.signal__send_no.emit()
        victim.SIGNALS.signal__send_no.emit()
        victim.SIGNALS.signal__send_no.emit()
        assert victim.value is None

    # -----------------------------------------------------------------------------------------------------------------
    def test__NO_PARAMS(self):
        victim = self.victim
        assert victim.value is None

        victim.value = 1
        assert victim.value == 1

        # CORRECT -----------------------------------------
        victim.SIGNALS.signal__send_no.emit()
        assert victim.value is None

        # INCORRECT -----------------------------------------
        victim.value = 1
        try:
            victim.SIGNALS.signal__send_no.emit(1)  # EXTRA
        except TypeError:
            assert True
        else:
            assert False

        assert victim.value == 1

    # -----------------------------------------------------------------------------------------------------------------
    def test__ONE(self):
        victim = self.victim
        assert victim.value is None

        # CORRECT -----------------------------------------------------------------------------------
        victim.SIGNALS.signal__send_one__wo_def.emit(1)
        assert victim.value == 1

        victim.SIGNALS.signal__send_one__wo_def__not_used.emit(11)
        assert victim.value == 999

        victim.SIGNALS.signal__send_one__with_def.emit(111)
        assert victim.value == 111

        # BLANK -----------------------------------------------------------------------------------
        # -----------------------------------------------
        try:
            victim.SIGNALS.signal__send_one__wo_def.emit()
        except TypeError:
            assert True
        else:
            assert False

        assert victim.value == 111

        # -----------------------------------------------
        try:
            victim.SIGNALS.signal__send_one__wo_def__not_used.emit()
        except TypeError:
            assert True
        else:
            assert False

        assert victim.value == 111

        # -----------------------------------------------
        try:
            victim.SIGNALS.signal__send_one__with_def.emit()        # ATTENTION!!! func woun't be called!
        except TypeError:
            assert True
        else:
            assert False

        assert victim.value == 111

        # WRONG TYPE -----------------------------------------------------------
        victim.SIGNALS.signal__send_one__wo_def.emit(True)
        # assert victim.value == int(True)
        assert victim.value == 1

        # in all such other cases always will be new different value!!!
        # WHY? dont matter! DECISION - DONT USE IT IN THIS WAY!!!

        # victim.SIGNALS.signal__send_one__wo_def.emit("1")
        # print(int("1"))     # 1
        # # assert victim.value == int("1")
        # # assert victim.value == 132102104
        # victim.SIGNALS.signal__send_one__wo_def.emit("2")
        # assert victim.value == 132102152                    # WHY????
        # victim.SIGNALS.signal__send_one__wo_def.emit("True")
        # assert victim.value == 132063408                    # WHY????
        # victim.SIGNALS.signal__send_one__wo_def.emit(None)
        # assert victim.value == 131245568                   # WHY????
        # # victim.SIGNALS.signal__send_one__wo_def.emit(0.1)
        # assert victim.value == -1428486704                   # WHY???? different value any time!!!

        # EXTRA COUNT -----------------------------------------------------------
        try:
            victim.SIGNALS.signal__send_one__wo_def.emit(11, 11)
        except TypeError:
            assert True
        else:
            assert False

    # -----------------------------------------------------------------------------------------------------------------
    def test__SEVERAL(self):
        victim = self.victim
        assert victim.value is None

        # CORRECT -----------------------------------------------------------
        victim.SIGNALS.signal__send_several__wo_def.emit(1, 2)
        assert victim.value == [1, 2]

        victim.SIGNALS.signal__send_several__with_def2.emit(11, 22)
        assert victim.value == [11, 22]

        victim.SIGNALS.signal__send_several__with_def12.emit(111, 222)
        assert victim.value == [111, 222]

        # INCORRECT -----------------------------------------------------------
        try:
            victim.SIGNALS.signal__send_several__wo_def.emit()                # LESS COUNT
        except TypeError:
            assert True
        else:
            assert False
        assert victim.value == [111, 222]

        try:
            victim.SIGNALS.signal__send_several__wo_def.emit(11)                # LESS COUNT
        except TypeError:
            assert True
        else:
            assert False
        assert victim.value == [111, 222]

        try:
            victim.SIGNALS.signal__send_several__wo_def.emit(11, 22, 33)     # EXTRA COUNT
        except TypeError:
            assert True
        else:
            assert False
        assert victim.value == [111, 222]


# =====================================================================================================================
